// lightmap.c

// ArghRad source - modified version of qrad3 source by id Software.
// ArghRad code modifications may not be used without permission
// from author Tim Wright. (Argh!)

// This version modified for Daikatana (dk)

#include "qrad.h"

// Argh!
extern	qboolean	minlimit;
extern	int			nocolor;
extern	qboolean	stopbleed;	// (1.50)
extern	qboolean	onlybounce;	// (2.00)
extern	qboolean	splotchfix; // (2.01)

// Argh! - sunlight data
#define MAX_SUNS 4

typedef struct
{
	char		*target;
	vec3_t		normal;
	vec3_t		color;
	double		light;
	double		diffuse;
	double		diffade;
	qboolean	use;
} sun_t;

sun_t	sun[MAX_SUNS];
float	sunambient;
vec3_t	sunambientcolor;
float	sunsurface;		// (2.00)
vec3_t	sunsurfacecolor;// (2.00)

/*
=================================================================

  POINT TRIANGULATION

=================================================================
*/

typedef struct triedge_s
{
	int			p0, p1;
	vec3_t		normal;
	vec_t		dist;
	struct triangle_s	*tri;
} triedge_t;

typedef struct triangle_s
{
	triedge_t	*edges[3];
} triangle_t;

#define	MAX_TRI_POINTS		2048 // (dk)
#define	MAX_TRI_EDGES		(MAX_TRI_POINTS*6)
#define	MAX_TRI_TRIS		(MAX_TRI_POINTS*2)

typedef struct
{
	int			numpoints;
	int			numedges;
	int			numtris;
	dplane_t	*plane;
	triedge_t	*edgematrix[MAX_TRI_POINTS][MAX_TRI_POINTS];
	patch_t		*points[MAX_TRI_POINTS];
	triedge_t	edges[MAX_TRI_EDGES];
	triangle_t	tris[MAX_TRI_TRIS];
} triangulation_t;

triangulation_t	*AllocTriangulation (dplane_t *plane)
{
	triangulation_t	*t;

	t = malloc(sizeof(triangulation_t));
	t->numpoints = 0;
	t->numedges = 0;
	t->numtris = 0;

	t->plane = plane;

	return t;
}

void FreeTriangulation (triangulation_t *tr)
{
	free (tr);
}


triedge_t	*FindEdge (triangulation_t *trian, int p0, int p1)
{
	triedge_t	*e, *be;
	vec3_t		v1;
	vec3_t		normal;
	vec_t		dist;

	if (trian->edgematrix[p0][p1])
		return trian->edgematrix[p0][p1];

	if (trian->numedges > MAX_TRI_EDGES-2)
		Error ("trian->numedges > MAX_TRI_EDGES-2");

	VectorSubtract (trian->points[p1]->origin, trian->points[p0]->origin, v1);
	VectorNormalize (v1, v1);
	CrossProduct (v1, trian->plane->normal, normal);
	dist = DotProduct (trian->points[p0]->origin, normal);

	e = &trian->edges[trian->numedges];
	e->p0 = p0;
	e->p1 = p1;
	e->tri = NULL;
	VectorCopy (normal, e->normal);
	e->dist = dist;
	trian->numedges++;
	trian->edgematrix[p0][p1] = e;

	be = &trian->edges[trian->numedges];
	be->p0 = p1;
	be->p1 = p0;
	be->tri = NULL;
	VectorSubtract (vec3_origin, normal, be->normal);
	be->dist = -dist;
	trian->numedges++;
	trian->edgematrix[p1][p0] = be;

	return e;
}

triangle_t	*AllocTriangle (triangulation_t *trian)
{
	triangle_t	*t;

	if (trian->numtris >= MAX_TRI_TRIS)
		Error ("trian->numtris >= MAX_TRI_TRIS");

	t = &trian->tris[trian->numtris];
	trian->numtris++;

	return t;
}

void TriEdge_r (triangulation_t *trian, triedge_t *e)
{
	int		i, bestp = 0;	// Knighmtare- init value
	vec3_t	v1, v2;
	vec_t	*p0, *p1, *p;
	vec_t	best, ang;
	triangle_t	*nt;

	if (e->tri)
		return;		// allready connected by someone

	// find the point with the best angle
	p0 = trian->points[e->p0]->origin;
	p1 = trian->points[e->p1]->origin;
	best = 1.1;
	for (i=0 ; i< trian->numpoints ; i++)
	{
		p = trian->points[i]->origin;
		// a 0 dist will form a degenerate triangle
		if (DotProduct(p, e->normal) - e->dist < 0)
			continue;	// behind edge
		VectorSubtract (p0, p, v1);
		VectorSubtract (p1, p, v2);
		if (!VectorNormalize (v1,v1))
			continue;
		if (!VectorNormalize (v2,v2))
			continue;
		ang = DotProduct (v1, v2);
		if (ang < best)
		{
			best = ang;
			bestp = i;
		}
	}
	if (best >= 1)
		return;		// edge doesn't match anything

	// make a new triangle
	nt = AllocTriangle (trian);
	nt->edges[0] = e;
	nt->edges[1] = FindEdge (trian, e->p1, bestp);
	nt->edges[2] = FindEdge (trian, bestp, e->p0);
	for (i=0 ; i<3 ; i++)
		nt->edges[i]->tri = nt;
	TriEdge_r (trian, FindEdge (trian, bestp, e->p1));
	TriEdge_r (trian, FindEdge (trian, e->p0, bestp));
}

void TriangulatePoints (triangulation_t *trian)
{
	vec_t	d, bestd;
	vec3_t	v1;
	int		bp1 = 0, bp2 = 0, i, j;	// Knighmtare- init values
	vec_t	*p1, *p2;
	triedge_t	*e, *e2;

	if (trian->numpoints < 2)
		return;

	// find the two closest points
	bestd = 999999;
	for (i=0 ; i<trian->numpoints ; i++)
	{
		p1 = trian->points[i]->origin;
		for (j=i+1 ; j<trian->numpoints ; j++)
		{
			p2 = trian->points[j]->origin;
			VectorSubtract (p2, p1, v1);
			d = VectorLength (v1);
			if (d < bestd)
			{
				bestd = d;
				bp1 = i;
				bp2 = j;
			}
		}
	}

	e = FindEdge (trian, bp1, bp2);
	e2 = FindEdge (trian, bp2, bp1);
	TriEdge_r (trian, e);
	TriEdge_r (trian, e2);
}

void AddPointToTriangulation (patch_t *patch, triangulation_t *trian)
{
	int			pnum;

	pnum = trian->numpoints;
	if (pnum == MAX_TRI_POINTS)
		Error ("trian->numpoints == MAX_TRI_POINTS");
	trian->points[pnum] = patch;
	trian->numpoints++;
}

void	LerpTriangle (triangulation_t *trian, triangle_t *t, vec3_t point, vec3_t color)
{
	patch_t		*p1, *p2, *p3;
	vec3_t		base, d1, d2;
	float		x, y, x1, y1, x2, y2;

	p1 = trian->points[t->edges[0]->p0];
	p2 = trian->points[t->edges[1]->p0];
	p3 = trian->points[t->edges[2]->p0];

	VectorCopy (p1->totallight, base);
	VectorSubtract (p2->totallight, base, d1);
	VectorSubtract (p3->totallight, base, d2);

	x = DotProduct (point, t->edges[0]->normal) - t->edges[0]->dist;
	y = DotProduct (point, t->edges[2]->normal) - t->edges[2]->dist;

	x1 = 0;
	y1 = DotProduct (p2->origin, t->edges[2]->normal) - t->edges[2]->dist;

	x2 = DotProduct (p3->origin, t->edges[0]->normal) - t->edges[0]->dist;
	y2 = 0;

	if (fabs(y1)<ON_EPSILON || fabs(x2)<ON_EPSILON)
	{
		VectorCopy (base, color);
		return;
	}

	VectorMA (base, x/x2, d2, color);
	VectorMA (color, y/y1, d1, color);
}


qboolean PointInTriangle (vec3_t point, triangle_t *t, int iscurve)
// Argh! - added iscurve to function params (2.00)
{
	int		i;
	triedge_t	*e;
	vec_t	d;

	for (i=0 ; i<3 ; i++)
	{
		e = t->edges[i];
		d = DotProduct (e->normal, point) - e->dist;

		// Argh! - tweak looks better on phong surfaces,
		//         but not on flat shaded surfaces (2.00)
		if (iscurve)
		{
			if (d <= 0)
				return false;	// not inside
		}
		else
		{
			if (d < 0)
				return false;	// not inside
		}
	}

	return true;
}

void SampleTriangulation (vec3_t point, triangulation_t *trian, int iscurve, vec3_t color)
// Argh! - added iscurve to function params (2.00)
{
	triangle_t	*t;
	triedge_t	*e;
	vec_t		d, best;
	patch_t		*p0, *p1;
	vec3_t		v1, v2;
	int			i, j;

	if (trian->numpoints == 0)
	{
		VectorClear (color);
		return;
	}
	if (trian->numpoints == 1)
	{
		VectorCopy (trian->points[0]->totallight, color);
		return;
	}

	// search for triangles
	for (t = trian->tris, j=0 ; j < trian->numtris ; t++, j++)
	{
		// Argh! - send iscurve too (2.00)
		if (!PointInTriangle (point, t, iscurve))
			continue;

		// this is it
		LerpTriangle (trian, t, point, color);
		return;
	}

	// search for exterior edge
	for (e=trian->edges, j=0 ; j< trian->numedges ; e++, j++)
	{
		if (e->tri)
			continue;		// not an exterior edge

		d = DotProduct (point, e->normal) - e->dist;
		if (d < 0)
			continue;	// not in front of edge

		p0 = trian->points[e->p0];
		p1 = trian->points[e->p1];

		VectorSubtract (p1->origin, p0->origin, v1);
		VectorNormalize (v1, v1);
		VectorSubtract (point, p0->origin, v2);
		d = DotProduct (v2, v1);
		if (d < 0)
			continue;
		if (d > 1)
			continue;
		for (i=0 ; i<3 ; i++)
			color[i] = p0->totallight[i] + d * (p1->totallight[i] - p0->totallight[i]);
		return;
	}

	// search for nearest point
	best = 999999;
	p1 = NULL;
	for (j=0 ; j<trian->numpoints ; j++)
	{
		p0 = trian->points[j];
		VectorSubtract (point, p0->origin, v1);
		d = VectorLength (v1);
		if (d < best)
		{
			best = d;
			p1 = p0;
		}
	}

	if (!p1)
		Error ("SampleTriangulation: no points");

	VectorCopy (p1->totallight, color);
}

/*
=================================================================

  LIGHTMAP SAMPLE GENERATION

=================================================================
*/


#define	SINGLEMAP	(64*64*4)

typedef struct
{
	vec_t	facedist;
	vec3_t	facenormal;

	int		numsurfpt;
	vec3_t	surfpt[SINGLEMAP];

	vec3_t	modelorg;		// for origined bmodels

	vec3_t	texorg;
	vec3_t	worldtotex[2];	// s = (world - texorg) . worldtotex[0]
	vec3_t	textoworld[2];	// world = texorg + s * textoworld[0]

	vec_t	exactmins[2], exactmaxs[2];

	int		texmins[2], texsize[2];
	int		surfnum;
	dface_t	*face;
} lightinfo_t;


/*
================
CalcFaceExtents

Fills in s->texmins[] and s->texsize[]
also sets exactmins[] and exactmaxs[]
================
*/
void CalcFaceExtents (lightinfo_t *l)
{
	dface_t *s;
	vec_t	mins[2], maxs[2], val;
	int		i,j, e;
	dvertex_t	*v;
	texinfo_t	*tex;
	vec3_t		vt;

	s = l->face;

	mins[0] = mins[1] = 999999;
	maxs[0] = maxs[1] = -999999;

	tex = &texinfo[s->texinfo];

	for (i=0 ; i<s->numedges ; i++)
	{
		e = dsurfedges[s->firstedge+i];
		if (e >= 0)
			v = dvertexes + dedges[e].v[0];
		else
			v = dvertexes + dedges[-e].v[1];

		VectorCopy (v->point, vt);

		for (j=0 ; j<2 ; j++)
		{
			val = DotProduct (vt, tex->vecs[j]) + tex->vecs[j][3];
			if (val < mins[j])
				mins[j] = val;
			if (val > maxs[j])
				maxs[j] = val;
		}
	}

	for (i=0 ; i<2 ; i++)
	{
		l->exactmins[i] = mins[i];
		l->exactmaxs[i] = maxs[i];

		mins[i] = floor(mins[i]/16);
		maxs[i] = ceil(maxs[i]/16);

		l->texmins[i] = mins[i];
		l->texsize[i] = maxs[i] - mins[i];
		if (l->texsize[0] * l->texsize[1] > SINGLEMAP/4)	// div 4 for extrasamples
			Error ("Surface to large to map");
	}
}

/*
================
CalcFaceVectors

Fills in texorg, worldtotex. and textoworld
================
*/
void CalcFaceVectors (lightinfo_t *l)
{
	texinfo_t	*tex;
	int			i, j;
	vec3_t	texnormal;
	vec_t	distscale;
	vec_t	dist, len;
	int			w, h;

	tex = &texinfo[l->face->texinfo];

// convert from float to double
	for (i=0 ; i<2 ; i++)
		for (j=0 ; j<3 ; j++)
			l->worldtotex[i][j] = tex->vecs[i][j];

// calculate a normal to the texture axis.  points can be moved along this
// without changing their S/T
	texnormal[0] = tex->vecs[1][1]*tex->vecs[0][2]
		- tex->vecs[1][2]*tex->vecs[0][1];
	texnormal[1] = tex->vecs[1][2]*tex->vecs[0][0]
		- tex->vecs[1][0]*tex->vecs[0][2];
	texnormal[2] = tex->vecs[1][0]*tex->vecs[0][1]
		- tex->vecs[1][1]*tex->vecs[0][0];
	VectorNormalize (texnormal, texnormal);

// flip it towards plane normal
	distscale = DotProduct (texnormal, l->facenormal);
	if (!distscale)
	{
		qprintf ("WARNING: Texture axis perpendicular to face\n");
		distscale = 1;
	}
	if (distscale < 0)
	{
		distscale = -distscale;
		VectorSubtract (vec3_origin, texnormal, texnormal);
	}

// distscale is the ratio of the distance along the texture normal to
// the distance along the plane normal
	distscale = 1/distscale;

	for (i=0 ; i<2 ; i++)
	{
		len = VectorLength (l->worldtotex[i]);
		dist = DotProduct (l->worldtotex[i], l->facenormal);
		dist *= distscale;
		VectorMA (l->worldtotex[i], -dist, texnormal, l->textoworld[i]);
		VectorScale (l->textoworld[i], (1/len)*(1/len), l->textoworld[i]);
	}


// calculate texorg on the texture plane
	for (i=0 ; i<3 ; i++)
		l->texorg[i] = -tex->vecs[0][3]* l->textoworld[0][i] - tex->vecs[1][3] * l->textoworld[1][i];

// project back to the face plane
	dist = DotProduct (l->texorg, l->facenormal) - l->facedist - 1;
	dist *= distscale;
	VectorMA (l->texorg, -dist, texnormal, l->texorg);

	// compensate for org'd bmodels
	VectorAdd (l->texorg, l->modelorg, l->texorg);

	// total sample count
	h = l->texsize[1]+1;
	w = l->texsize[0]+1;
	l->numsurfpt = w * h;
}

/*
=================
CalcPoints

For each texture aligned grid point, back project onto the plane
to get the world xyz value of the sample point
=================
*/
void CalcPoints (lightinfo_t *l, float sofs, float tofs)
{
	int		i;
	int		s, t, j;
	int		w, h, step;
	vec_t	starts, startt, us, ut;
	vec_t	*surf;
	vec_t	mids, midt;
	vec3_t	facemid;
	dleaf_t	*leaf;

	surf = l->surfpt[0];
	mids = (l->exactmaxs[0] + l->exactmins[0])/2;
	midt = (l->exactmaxs[1] + l->exactmins[1])/2;

	for (j=0 ; j<3 ; j++)
		facemid[j] = l->texorg[j] + l->textoworld[0][j]*mids + l->textoworld[1][j]*midt;

	h = l->texsize[1]+1;
	w = l->texsize[0]+1;
	l->numsurfpt = w * h;

	starts = l->texmins[0]*16;
	startt = l->texmins[1]*16;
	step = 16;


	for (t=0 ; t<h ; t++)
	{
		for (s=0 ; s<w ; s++, surf+=3)
		{
			us = starts + (s+sofs)*step;
			ut = startt + (t+tofs)*step;


		// if a line can be traced from surf to facemid, the point is good
			for (i=0 ; i<6 ; i++)
			{
			// calculate texture point
				for (j=0 ; j<3 ; j++)
					surf[j] = l->texorg[j] + l->textoworld[0][j]*us
					+ l->textoworld[1][j]*ut;

				leaf = PointInLeaf (surf);
				if (leaf->contents != CONTENTS_SOLID)
				{
					if (!TestLine_r (0, facemid, surf))
						break;	// got it
				}

				// nudge it
				if (i & 1)
				{
					if (us > mids)
					{
						us -= 8;
						if (us < mids)
							us = mids;
					}
					else
					{
						us += 8;
						if (us > mids)
							us = mids;
					}
				}
				else
				{
					if (ut > midt)
					{
						ut -= 8;
						if (ut < midt)
							ut = midt;
					}
					else
					{
						ut += 8;
						if (ut > midt)
							ut = midt;
					}
				}
			}
		}
	}

}


//==============================================================



#define	MAX_STYLES	32
typedef struct
{
	int			numsamples;
	float		*origins;
	int			numstyles;
	int			stylenums[MAX_STYLES];
	float		*samples[MAX_STYLES];
} facelight_t;

directlight_t	*directlights[MAX_MAP_LEAFS];
facelight_t		facelight[MAX_MAP_FACES];
int				numdlights;
int				numdlights2; // (dk)

/*
==================
FindTargetEntity
==================
*/
entity_t *FindTargetEntity (char *target)
{
	int		i;
	char	*n;

	for (i=0 ; i<num_entities ; i++)
	{
		n = ValueForKey (&entities[i], "targetname");
		if (!strcmp (n, target))
			return &entities[i];
	}

	return NULL;
}

#define	DIRECT_LIGHT	1

/*
=============
CreateDirectLightsFromEntities
=============
*/
void CreateDirectLightsFromEntities (void)
{
	int		i;
	directlight_t	*dl;
	dleaf_t	*leaf;
	int		cluster;
	entity_t	*e, *e2;
	char	*name;
	char	*target;
	float	angle;
	vec3_t	dest;
	char	*_color;
	float	intensity;
	// Argh!
	char	*spotvec;	// spotlight vector
	char	*spotang;	// spotlight angles
	char	*spotpnt;	// spotlight point
	char	*fade;		// pointlight dist fade
	char	*afade;		// pointlight angle fade
	char	*focus;		// directional focus fade
	char	*falloff;	// pointlight falloff type
	float	coneValue;	// (dk)

	//
	// entities
	//
	for (i=0 ; i<num_entities ; i++)
	{
		e = &entities[i];
		name = ValueForKey (e, "classname");

		// Argh! - get sun data from worldspawn
		if (!strcmp(name, "worldspawn"))
		{
			float value;
			char *amb;
			int s;

			// go through all suns
			for (s = 0; s < MAX_SUNS; s++)
			{
				char key[32];

				// init all sun fields
				sun[s].light = 200.0;	// default brightness
				sun[s].diffuse = 0.0;	// 0 diffuse brightness
				sun[s].diffade = 1.0;	// standard diffuse fade
				sun[s].color[0] = 0.0;	// no color
				sun[s].color[1] = 0.0;
				sun[s].color[2] = 0.0;
				sun[s].target = NULL;	// no targetname
				sun[s].normal[0] = 0.0;	// point sun straight down
				sun[s].normal[1] = 0.0;
				sun[s].normal[2] = -1.0;
				sun[s].use = FALSE;		// not in use

				// get sun brightness
				if (s == 0)
					sprintf(key, "_sun_light");
				else
					sprintf(key, "_sun%u_light", s + 1);
				value = FloatForKey(e, key);
				if (value)
				{
					sun[s].light = value;
					sun[s].use = TRUE;
				}

				// get sun diffuse
				if (s == 0)
					sprintf(key, "_sun_diffuse");
				else
					sprintf(key, "_sun%u_diffuse", s + 1);
				value = FloatForKey(e, key);
				if (value)
				{
					if (value < sun[s].light)
						sun[s].diffuse = value;
					else
						sun[s].diffuse = sun[s].light;
					sun[s].use = TRUE;
				}

				// get sun diffuse fade
				if (s == 0)
					sprintf(key, "_sun_difwait");
				else
					sprintf(key, "_sun%u_difwait", s + 1);
				value = FloatForKey(e, key);
				if (!value)
				{
					if (s == 0)
						sprintf(key, "_sun_diffade");
					else
						sprintf(key, "_sun%u_diffade", s + 1);
					value = FloatForKey(e, key);
				}
				if (value)
					sun[s].diffade = value;

				// get sun color
				if (s == 0)
					sprintf(key, "_sun_color");
				else
					sprintf(key, "_sun%u_color", s + 1);
				GetVectorForKey(e, key, sun[s].color);
				if (sun[s].color[0] || sun[s].color[1] || sun[s].color[2])
				{
					VectorNormalize(sun[s].color, sun[s].color);
					sun[s].use = TRUE;
				}

				// get sun target - spotlight targeted sunlight
				if (s == 0)
					sprintf(key, "_sun");
				else
					sprintf(key, "_sun%u", s + 1);
				sun[s].target = ValueForKey(e, key);
				if (!sun[s].target[0])
				{
					if (s == 0)
						sprintf(key, "_sun_target");
					else
						sprintf(key, "_sun%u_target", s + 1);
					sun[s].target = ValueForKey(e, key);
				}
				if (sun[s].target[0])
				{
					sun[s].use = TRUE;
				}
				else // get sun mangle - pitch/yaw angled sunlight
				{
					char *mangle;

					if (s == 0)
						sprintf(key, "_sun_mangle");
					else
						sprintf(key, "_sun%u_mangle", s + 1);
					mangle = ValueForKey(e, key);
					if (!mangle[0])
					{
						if (s == 0)
							sprintf(key, "_sun_angle");
						else
							sprintf(key, "_sun%u_angle", s + 1);
						mangle = ValueForKey(e, key);
					}

					if (mangle[0])
					{
						float yaw, pitch;

						sscanf (mangle, "%f %f", &yaw, &pitch);
						sun[s].normal[0] = cos(yaw * Q_PI / 180) * (sun[s].normal[1] = cos(pitch * Q_PI / 180));
						sun[s].normal[1] *= sin(yaw * Q_PI / 180);
						sun[s].normal[2] = sin(pitch * Q_PI / 180);
						VectorNormalize (sun[s].normal, sun[s].normal);
						sun[s].use = TRUE;
					}
					else // get sun vector - vector directed sunlight
					{
						if (s == 0)
							sprintf(key, "_sun_vector");
						else
							sprintf(key, "_sun%u_vector", s + 1);
						GetVectorForKey(e, key, sun[s].normal);

						if (sun[s].normal[0] || sun[s].normal[1] || sun[s].normal[2])
						{
							VectorNormalize(sun[s].normal, sun[s].normal);
							sun[s].use = TRUE;
						}
					}
				}
			}

			// get sun ambient (rev2.00)
			sunambient = 0.0;
			amb = ValueForKey(e, "_sun_ambient");
			if (amb[0])
			{
				int n = sscanf(amb, "%f %f %f", &sunambientcolor[0], &sunambientcolor[1], &sunambientcolor[2]);
				if (n < 3)
				{
					sunambient = sunambientcolor[0];
					sunambientcolor[0] = 0;
					sunambientcolor[1] = 0;
				}
				else
					sunambient = ColorNormalize(sunambientcolor, sunambientcolor);
			}

			// get sun surface (2.00)
			sunsurface = 0.0;
			amb = ValueForKey(e, "_sun_surface");
			if (amb[0])
			{
				int n = sscanf(amb, "%f %f %f", &sunsurfacecolor[0], &sunsurfacecolor[1], &sunsurfacecolor[2]);
				if (n < 3)
				{
					sunsurface = sunsurfacecolor[0];
					sunsurfacecolor[0] = 0;
					sunsurfacecolor[1] = 0;
				}
				else
					sunsurface = ColorNormalize(sunsurfacecolor, sunsurfacecolor);
			}

			// get global ambient light
			amb = ValueForKey(e, "ambient"); // (dk)
			if (!amb[0])
				amb = ValueForKey(e, "light");
			if (!amb[0])
				amb = ValueForKey(e, "_minlight");
			if (!amb[0])
				amb = ValueForKey(e, "_ambient");
			if (amb[0])
			{
				int n = sscanf(amb, "%f %f %f", &ambient[0], &ambient[1], &ambient[2]);
				if (n < 3)
				{
					ambient[1] = ambient[0];
					ambient[2] = ambient[0];
				}
			}
		}

		if (strncmp (name, "light", 5))
			continue;

		// numdlights++;
		dl = malloc(sizeof(directlight_t));
		memset (dl, 0, sizeof(*dl));

		dl->face = NULL; // Argh! - no face for ents (2.00)

#if 0
		// set bounce flag (dk)
		if (FloatForKey(e, "bounce"))
		{
			numdlights++;
			dl->bounce = true;
		}
		else
		{
			numdlights2++;
			dl->bounce = false;
		}
#else
		// Knightmare- make all lights bounce
		numdlights++;
		dl->bounce = true;
#endif

		// set max brightness cap (dk)
		dl->cap = FloatForKey (e, "cap");

		GetVectorForKey (e, "origin", dl->origin);
		dl->style = FloatForKey (e, "_style");
		if (!dl->style)
			dl->style = FloatForKey (e, "style");
		if (dl->style < 0 || dl->style >= MAX_LSTYLES)
			dl->style = 0;

		leaf = PointInLeaf (dl->origin);
		cluster = leaf->cluster;

		dl->next = directlights[cluster];
		directlights[cluster] = dl;

		intensity = FloatForKey (e, "light");
		if (!intensity)
			intensity = FloatForKey (e, "_light");
		if (!intensity)
			intensity = 300;
		_color = ValueForKey (e, "_color");
		if (_color && _color[1])
		{
			sscanf (_color, "%f %f %f", &dl->color[0],&dl->color[1],&dl->color[2]);
			// Argh! - gonna keep this in for now (dk)
			// Nelno: don't adjust light saturation?
			ColorNormalize (dl->color, dl->color);
		}
		else
			dl->color[0] = dl->color[1] = dl->color[2] = 1.0;
		dl->intensity = intensity*entity_scale;
		dl->type = emit_point;

		// Argh! - next few blocks use ValueForKey
		//         to handle a value of 0 properly

		// Argh! - store distance fade
		//         added "falloff" key (dk)
		fade = ValueForKey (e, "falloff");
		if (fade[0])
			dl->fade = atof(fade);
		else
		{
			fade = ValueForKey (e, "_fade");
			if (fade[0])
				dl->fade = atof(fade);
			else
			{
				fade = ValueForKey (e, "wait");
				if (fade[0])
					dl->fade = atof(fade);
				else
				{
					fade = ValueForKey (e, "_wait");
					if (fade[0])
						dl->fade = atof(fade);
					else
						dl->fade = 1.0; // default
				}
			}
		}
		if (dl->fade < 0.0)
			dl->fade = 0.0;

		// Argh! - store angle fade
		afade = ValueForKey (e, "_angfade");
		if (afade[0])
			dl->afade = atof(afade);
		else
		{
			afade = ValueForKey (e, "_angwait");
			if (afade[0])
				dl->afade = atof(afade);
			else
				dl->afade = 1.0; // default
		}
		if (dl->afade < 0.0)
			dl->afade = 0.0;

		// Argh! - store focus fade
		focus = ValueForKey (e, "_focus");
		if (focus[0])
			dl->focus = atof(focus);
		else
			dl->focus = 1.0;
		if (dl->focus < 0.0)
			dl->focus = 0.0;

		// Argh! - store falloff type
		falloff = ValueForKey(e, "_falloff");
		if (falloff[0])
			dl->falloff = atoi(falloff);
		else
			dl->falloff = 0;
		if ((dl->falloff < 0) || (dl->falloff > 2))
			dl->falloff = 0;

		// Argh! - spotlight code rearranged for new keys
		target = ValueForKey (e, "target");
		spotang = ValueForKey (e, "_spotangle");
		if (!spotang[0])
			spotang = ValueForKey (e, "_mangle");
		spotvec = ValueForKey (e, "_spotvector");
		spotpnt = ValueForKey (e, "_spotpoint");
		// AMW:  Sverre wants the "cone/_cone" key to work on all lights (dk)
		coneValue = FloatForKey (e, "cone");
		if (coneValue == 0.0)
			coneValue = FloatForKey (e, "_cone");

		if (!strcmp (name, "light_spot") || target[0] ||
			spotvec[0] || spotang[0] || spotpnt[0] || coneValue)
		{
			dl->type = emit_spotlight;

			dl->stopdot = coneValue; // (dk)
			if (!dl->stopdot)
				dl->stopdot = 10;
			dl->stopdot = cos(dl->stopdot/180*3.14159);

			if (target[0]) // entity-targeted lights
			{
				// point towards target
				e2 = FindTargetEntity (target);
				if (!e2)
				{
					printf ("WARNING: light at (%i %i %i) has missing target\n",
					        (int)dl->origin[0], (int)dl->origin[1], (int)dl->origin[2]);
					dl->normal[2] = 0;
					dl->normal[0] = 1;
					dl->normal[1] = 0;
				}
				else
				{
					int s; // Argh - for sun

					GetVectorForKey (e2, "origin", dest);
					VectorSubtract (dest, dl->origin, dl->normal);
					VectorNormalize (dl->normal, dl->normal);

					// check for sunlight target (fixed1.01)
					for (s = 0; s < MAX_SUNS; s++)
					{
						if (sun[s].use && !strcmp(target, sun[s].target))
							VectorCopy(dl->normal, sun[s].normal);
					}

				}
			}
			else if (spotang[0]) // Argh! - pitch/yaw angled lights
			{
				float yaw, pitch;

				sscanf (spotang, "%f %f", &yaw, &pitch);
				dl->normal[0] = cos(yaw * Q_PI / 180) * (dl->normal[1] = cos(pitch * Q_PI / 180));
				dl->normal[1] *= sin(yaw * Q_PI / 180);
				dl->normal[2] = sin(pitch * Q_PI / 180);
				VectorNormalize (dl->normal, dl->normal);
			}
			else if (spotvec[0]) // Argh! - vector directed lights
			{
				GetVectorForKey (e, "_spotvector", dest);
				VectorCopy(dest, dl->normal);
				VectorNormalize (dl->normal, dl->normal);
			}
			else if (spotpnt[0]) // Argh! - point targeted lights
			{
				GetVectorForKey (e, "_spotpoint", dest);
				VectorSubtract (dest, dl->origin, dl->normal);
				VectorNormalize (dl->normal, dl->normal);
			}
			else // rearranged again (dk)
			{
				angle = FloatForKey (e, "angle");
				if (angle == ANGLE_UP)
				{
					dl->normal[0] = dl->normal[1] = 0;
					dl->normal[2] = 1;
				}
				else if (angle == ANGLE_DOWN)
				{
					dl->normal[0] = dl->normal[1] = 0;
					dl->normal[2] = -1;
				}
				else
				{
					dl->normal[2] = 0;
					dl->normal[0] = cos (angle/180*3.14159);
					dl->normal[1] = sin (angle/180*3.14159);
				}
			}
		}

	}
}

void CreateDirectLightsFromSurfaces (void)
{
	int		i;
	patch_t	*p;
	directlight_t	*dl;
	dleaf_t	*leaf;
	int		cluster;

	for (i=0, p=patches ; i<num_patches ; i++, p++)
	{
		// Argh! - vars for sun
		qboolean usesun = false;
		int s;

		// Argh! - fabs accounts for antilights
		if (fabs(p->totallight[0]) < DIRECT_LIGHT
			&& fabs(p->totallight[1]) < DIRECT_LIGHT
			&& fabs(p->totallight[2]) < DIRECT_LIGHT)
			continue;

		numdlights++;
		dl = malloc(sizeof(directlight_t));
		memset (dl, 0, sizeof(*dl));

		VectorCopy (p->origin, dl->origin);

		leaf = PointInLeaf (dl->origin);
		cluster = leaf->cluster;
		dl->next = directlights[cluster];
		directlights[cluster] = dl;

		// Argh! - set sunlight type (rev2.00)
		for (s = 0; s < MAX_SUNS; s++)
			usesun |= sun[s].use;
		if (IsSky(p->face) && (usesun || sunambient || sunsurface))
			dl->type = emit_sunlight;
		else
			dl->type = emit_surface;

		VectorCopy (p->plane->normal, dl->normal);

		// Argh! - copy settings to light
		dl->face  = p->face; // (2.00)
		dl->focus = p->focus;
		dl->style = p->style;

		dl->bounce = true; // all surface lights bounce (dk)

		dl->intensity = ColorNormalize (p->totallight, dl->color);
		dl->intensity *= p->area * direct_scale;
		VectorClear (p->totallight);	// all sent now
	}
}

// reorganized into multiple functions (dk)
void CreateDirectLights (void)
{
	numdlights  = 0;
	numdlights2 = 0;

	// need worldspawn settings so run entities first
	CreateDirectLightsFromEntities ();

	CreateDirectLightsFromSurfaces ();

	qprintf ("%i direct lights - with bounce\n", numdlights);
	qprintf ("%i direct lights - no bounce\n", numdlights2);
}



/*
=============
GatherSampleLight

Lightscale is the normalizer for multisampling
=============
*/
#define	DIFFADE_DIST	128 // Argh! - (2.00)

// added bouncelight to args (dk)
void GatherSampleLight (vec3_t pos, vec3_t normal, float **styletable,
	float *bouncelight, int offset, int mapsize, float lightscale)
{
	int				i;
	directlight_t	*l;
	byte			pvs[(MAX_MAP_LEAFS+7)/8];
	vec3_t			delta;
	float			dot, dot2;
	float			dist;
	float			scale = 0;	// Knightmare- default to 0
	float			*dest;
	// Argh - (rev2.00)
	vec3_t		delta2, antisun, skypos;
	float		dot3, dist2;
	float		sunlight[MAX_SUNS];	// direct  sunlight scale
	float		sundiff[MAX_SUNS];	// diffuse sunlight scale
	float		sunamb;				// ambient sunlight scale
	directlight_t	*sunlightl[MAX_SUNS];	// ent used for direct sunlight
	directlight_t	*sundiffl[MAX_SUNS];	// ent used for diffuse sunlight
	directlight_t	*sunambl;				// ent used for ambient sunlight
	qboolean	tl;			// true if testline computed (2.00)
	int			testline = 0;	// line test from pos to l->pos (2.00)	// Knightmare- default to 0
	int			s;
	int			numclusters;// depends on visdata
	float		subd;		// (2.00)
	float		cutoff;		// (2.00)
	float		*dest2;		// for bouncing light (dk)

	// Argh! - init sun data (rev2.00)
	sunambl = NULL;
	sunamb  = 0.0;
	for (s = 0; s < MAX_SUNS; s++)
	{
		sunlightl[s] = NULL;
		sundiffl[s]  = NULL;
		sunlight[s] = 0.0;
		sundiff[s]  = 0.0;
	}

	// get the PVS for the pos to limit the number of checks
	if (!PvsForOrigin (pos, pvs))
	{
		return;
	}

	// Argh! - don't use visdata value if it doesn't exist!,
	//         enables direct lighting in unvised maps (rev2.00)
	if (visdatasize)
		numclusters = dvis->numclusters;
	else
		numclusters = 8;

	for (i = 0 ; i<numclusters ; i++)
	{
		if ( !(pvs[ i>>3] & (1<<(i&7))) )
			continue;

		for (l=directlights[i] ; l ; l=l->next)
		{
			VectorSubtract (l->origin, pos, delta);
			dist = VectorNormalize (delta, delta);
			dot = DotProduct (delta, normal);
			if (dot < 0.001)
				continue;	// behind sample surface

			// Argh! - testline not yet computed for this l (2.00)
			tl = FALSE;

			switch (l->type)
			{
			case emit_point: // linear falloff (changable)

				// Argh! - factor in distance fade
				if (l->falloff == 0)
				{
					if (l->fade != 1.0)
						dist = dist * l->fade;

					if (dist > fabs(l->intensity))
						continue; // outside falloff
				}

				// Argh! - factor in angle fade
				if (l->afade != 1.0)
					dot = (1.0 - l->afade) + l->afade * dot;

				// Argh! - account for falloff
				if (l->falloff == 0)
				{
					// Argh! - account for antilights
					if (l->intensity >= 0.0)
						scale = (l->intensity - dist) * dot;
					else
						scale = (l->intensity + dist) * dot;
				}
				else if (l->falloff == 1)
				{
					scale = (l->intensity / dist) * dot;
				}
				else
				{
					scale = (l->intensity / (dist*dist)) * dot;
				}
				break;

			case emit_surface: // inverse square falloff

				dot2 = -DotProduct (delta, l->normal);

				// Argh! - set subdiv* and cutoff (2.00)
				if (IsSky(l->face))
				{
					cutoff = -0.1; // allow slightly behind light surface
					subd = subdivsky;
				}
				else if (IsWarp(l->face))
				{
					cutoff = -0.1; // allow slightly behind light surface
					subd = subdivwarp;
				}
				else
				{
					cutoff = 0.001; // behind light surface
					if (splotchfix) // can disable for plain lights (2.01)
						subd = subdivlight;
					else
						subd = 0;
				}

				if (dot2 < cutoff)
					continue;

				// Argh! - helps eliminate black-band
				//         and splotch problems (2.00)
				if (dist < subd)
				{
					dot  = dot  * dist/subd + (subd - dist)/subd;
					dot2 = dot2 * dist/subd + (subd - dist)/subd * 2;
					dist = subd;
				}

				// Argh - factor in focus
				if (l->focus != 1.0)
					dot2 = (1.0 - l->focus) + l->focus * dot2;

				scale = (l->intensity / (dist*dist)) * dot * dot2;

				break;

			case emit_spotlight: // linear falloff (changable)

				// Argh! - factor in distance fade
				if (l->falloff == 0)
				{
					if (l->fade != 1.0)
						dist = dist * l->fade;

					if (dist > fabs(l->intensity))
						continue; // outside falloff
				}

				dot2 = -DotProduct (delta, l->normal);
				if (dot2 <= l->stopdot)
					continue; // outside light cone

				// Argh - factor in focus (fixed1.02)
				if (l->focus != 1.0)
				{
					dot2 = (1.0 - l->focus) + l->focus * dot2;

					if (dot2 < 0)
						continue; // outside focus
				}

				// Argh! - factor in angle fade
				if (l->afade != 1.0)
					dot = (1.0 - l->afade) + l->afade * dot;

				// Argh! - account for antilights
				if (l->falloff == 0)
				{
					if (l->intensity >= 0.0)
						scale = (l->intensity - dist) * dot * dot2;
					else
						scale = (l->intensity + dist) * dot * dot2;
				}
				else if (l->falloff == 1)
				{
					scale = (l->intensity / dist) * dot * dot2;
				}
				else
				{
					scale = (l->intensity / (dist*dist)) * dot * dot2;
				}
				break;

			// Argh! - sunlight options
			case emit_sunlight: // no falloff

				// direct/diffuse/ambient sunlight are now kept
				// track of but not added to the lightmap until
				// all other lights are finished, surface sunlight
				// is added as part of the normal loop (rev2.00)

				dot2 = -DotProduct (delta, l->normal);

				// allow slightly behind light surface (2.00)
				if (dot2 < -0.1)
					continue;

				// surface sunlight (2.00)
				if (sunsurface)
				{
					// helps eliminate black-band and splotch problems
					if (dist < subdivsky)
					{
						dot  = dot  * dist/subdivsky + (subdivsky - dist)/subdivsky;
						dot2 = dot2 * dist/subdivsky + (subdivsky - dist)/subdivsky * 2;
						dist = subdivsky;
					}

					if (sunsurface <= 1)
						scale = (l->intensity / (dist*dist)) * dot * dot2;
					else
					{
						float temp;

						// scale sunsurface intensity to same
						// % of total surface light that l is
						temp = sunsurface * l->intensity / texinfo[l->face->texinfo].value;
						scale = (temp / (dist*dist)) * dot * dot2;
					}
				}
				else
					scale = 0.0;

				// go through all suns
				for (s = 0; s < MAX_SUNS; s++)
				{
					if (!sun[s].use)
						continue; // not used

					if (sunlightl[s])
						continue; // already cast (2.00)

					dot3 = -DotProduct(normal, sun[s].normal);
					if (dot3 <= 0.001) // behind sun direction
						continue;

					// set new dist
					VectorScale(sun[s].normal, -1, antisun);
					VectorSubtract(l->origin, pos, delta); // unnormalized delta
					dist = DotProduct(l->normal, delta) / DotProduct(l->normal, antisun);

					// find sky intersection point
					VectorScale(antisun, dist, antisun);
					VectorAdd(antisun, pos, skypos);

					// direct sunlight (rev 2.00)
					// only do lighting if within limited area
					VectorSubtract (l->origin, skypos, delta2);
					dist2 = VectorNormalize (delta2, delta2);

					if ((dist2 / 2.0) <= subdivsky)
					{
						// if not blocked, set sunlight
						if (!TestLine_r (0, pos, skypos) &&
							!TestLine_r (0, l->origin, skypos))
						{
							sunlight[s] = sun[s].light * dot3;
							sunlightl[s] = l;
						}
					}

					// diffuse sunlight (rev2.00)
					if (sun[s].diffuse && !sunlightl[s])
					{
						// compute testline if not previously done
						if (!tl)
						{
							testline = TestLine_r (0, pos, l->origin);
							tl = TRUE;
						}

						if (!testline)
						{
							// factor in distance fade
							if (sun[s].diffade != 1.0)
								dist2 = dist2 * sun[s].diffade;

							// if within DIFFADE_DIST, set diffuse
							if (dist2 < DIFFADE_DIST)
							{
								float diff = (DIFFADE_DIST - dist2) / DIFFADE_DIST * sun[s].diffuse * dot3;

								// if new value brighter, use it
								if (sundiff[s] < diff)
								{
									sundiff[s] = diff;
									sundiffl[s] = l;
								}
							}
						}

					}

				}

				// ambient sunlight (rev2.00)
				if (sunambient && !sunambl)
				{
					// compute testline if not previously done
					if (!tl)
					{
						testline = TestLine_r (0, pos, l->origin);
						tl = TRUE;
					}

					if (!testline)
					{
						sunamb = sunambient;
						sunambl = l;
					}
				}
				break;

			default:
				Error ("Bad l->type");
			}

			// Argh! - add in lighting from this light,
			//         rearranged for sunlight (rev2.00)

			if (scale == 0)
				continue; // no light, move on

			// Argh! - compute testline if not previously done (2.00)
			if (!tl)
				testline = TestLine_r (0, pos, l->origin);
			if (testline)
				continue; // occluded, move on

			// max brightness cap (dk)
			if (l->cap)
			{
				if (scale > l->cap)
					scale = l->cap;
			}
			
			// if this style doesn't have a table yet, allocate one
			if (!styletable[l->style])
			{
				styletable[l->style] = malloc (mapsize);
				memset (styletable[l->style], 0, mapsize);
			}

			// add some light to it
			dest  = styletable[l->style] + offset;
			dest2 = bouncelight + offset; // (dk)

			if (l->type != emit_sunlight) // non-sunlight
			{
				VectorMA (dest, scale*lightscale, l->color, dest);

				// only add style-0 bouncing lights to bouncelight (dk)
				if (l->bounce && !l->style)
					VectorMA (dest2, scale*lightscale, l->color, dest2);
			}
			else // Argh! - set sunsurface color (2.00)
			{
				// sunlight always bounces (dk)

				// use sunsurface color if available
				if (sunsurfacecolor[0] || sunsurfacecolor[1] || sunsurfacecolor[2])
				{
					VectorMA (dest,  scale*lightscale, sunsurfacecolor, dest);
					VectorMA (dest2, scale*lightscale, sunsurfacecolor, dest2);
				}
				// otherwise, compute based on average sun/light colors
				else
				{
					vec3_t temp;

					VectorClear(temp);
					for (s = 0; s < MAX_SUNS; s++)
					{
						if (!sun[s].use)
							continue; // unused

						// use sun color if available
						if (sun[s].color[0] || sun[s].color[1] || sun[s].color[2])
							VectorMA(temp, sun[s].light, sun[s].color, temp);
						// otherwise use light color
						else
							VectorMA(temp, sun[s].light, l->color, temp);
					}
					ColorNormalize(temp, temp);
					VectorMA (dest,  scale*lightscale, temp, dest);
					VectorMA (dest2, scale*lightscale, temp, dest2);
				}
			}

		}
	}

	// Argh! - all normal lights processed, now add in
	//         the direct/diffuse/ambient sunlight (rev2.00)

	// alloc table if needed
	if (!styletable[0])
	{
		styletable[0] = malloc (mapsize);
		memset (styletable[0], 0, mapsize);
	}

	dest  = styletable[0] + offset;
	dest2 = bouncelight + offset; // sunlight always bounces (dk)

	// direct/diffuse sunlight
	for (s = 0; s < MAX_SUNS; s++)
	{
		if (!sun[s].use)
			continue; // unused

		// direct
		if (sunlightl[s])
		{
			// use sun color if available
			if (sun[s].color[0] || sun[s].color[1] || sun[s].color[2])
			{
				VectorMA(dest,  sunlight[s]*lightscale, sun[s].color, dest);
				VectorMA(dest2, sunlight[s]*lightscale, sun[s].color, dest2);
			}
			// otherwise, use light color
			else
			{
				VectorMA(dest,  sunlight[s]*lightscale, sunlightl[s]->color, dest);
				VectorMA(dest2, sunlight[s]*lightscale, sunlightl[s]->color, dest2);
			}
		}
		// diffuse
		else if (sundiffl[s])
		{
			// use sun color if available
			if (sun[s].color[0] || sun[s].color[1] || sun[s].color[2])
			{
				VectorMA(dest,  sundiff[s]*lightscale, sun[s].color, dest);
				VectorMA(dest2, sundiff[s]*lightscale, sun[s].color, dest2);
			}
			// otherwise, use light color
			else
			{
				VectorMA(dest,  sundiff[s]*lightscale, sundiffl[s]->color, dest);
				VectorMA(dest2, sundiff[s]*lightscale, sundiffl[s]->color, dest2);
			}
		}
	}

	// ambient sunlight
	if (sunambl)
	{
		// use ambient color if available
		if (sunambientcolor[0] || sunambientcolor[1] || sunambientcolor[2])
		{
			VectorMA (dest,  sunamb*lightscale, sunambientcolor, dest);
			VectorMA (dest2, sunamb*lightscale, sunambientcolor, dest2);
		}
		// otherwise, compute based on sun/light colors
		else
		{
			vec3_t temp;

			VectorClear(temp);
			for (s = 0; s < MAX_SUNS; s++)
			{
				if (!sun[s].use)
					continue; // unused

				// use sun color if available
				if (sun[s].color[0] || sun[s].color[1] || sun[s].color[2])
					VectorMA(temp, sun[s].light, sun[s].color, temp);
				// otherwise use light color
				else
					VectorMA(temp, sun[s].light, sunambl->color, temp);
			}

			// use avaraged sunlight if available
			if (temp[0] || temp[1] || temp[2])
			{
				ColorNormalize(temp, temp);
				VectorMA (dest,  sunamb*lightscale, temp, dest);
				VectorMA (dest2, sunamb*lightscale, temp, dest2);
			}
			// otherwise use light color
			else
			{
				VectorMA (dest,  sunamb*lightscale, sunambl->color, dest);
				VectorMA (dest2, sunamb*lightscale, sunambl->color, dest2);
			}
		}
	}

}


/* from SiN qrad3 util (1.50)
=============
FreeDirectLights
=============
*/
void FreeDirectLights ( void )
{
	int i;
	directlight_t *l, *next;

	for (i = 0 ; i<MAX_MAP_LEAFS ; i++)
	{
		for (l=directlights[i] ; l ; l=next)
		{
			next = l->next;
			free(l);
		}
	}
}

/*
=============
AddSampleToPatch

Take the sample's collected light and
add it back into the apropriate patch
for the radiosity pass.

The sample is added to all patches that might include
any part of it.  They are counted and averaged, so it
doesn't generate extra light.
=============
*/
void AddSampleToPatch (vec3_t pos, vec3_t color, int facenum)
{
	patch_t	*patch;
	vec3_t	mins, maxs;
	int		i;

	if (numbounce == 0)
		return;
	if (color[0] + color[1] + color[2] < 3)
		return;

	for (patch = face_patches[facenum] ; patch ; patch=patch->next)
	{
		// see if the point is in this patch (roughly)
		WindingBounds (patch->winding, mins, maxs);

		for (i=0 ; i<3 ; i++)
		{
			if (mins[i] > pos[i] + 16)
				goto nextpatch;
			if (maxs[i] < pos[i] - 16)
				goto nextpatch;
		}

		// add the sample to the patch
		patch->samples++;
		VectorAdd (patch->samplelight, color, patch->samplelight);
nextpatch:;
	}

}


// Argh - retrieves the normal at the specified
//        vertex for the specified curve (2.00)
void GetVertexNormal(int vert, int tag, vec3_t normal)
{
	vnorms_t *vn;

	for (vn=vertnorms[vert]; vn; vn=vn->next)
	{
		if (vn->tag == tag)
		{
			VectorCopy(vn->normal, normal);
			return;
		}
	}

	// only happens if wrong info given
	VectorClear(normal);
	return;
}


// Argh - calculates normal for given face point,
//        performs phong calculations if needed. (2.00)
void GetPhongNormal (dface_t *f, vec3_t pnt, vec3_t oldnormal, vec3_t normal)
{
	int		a1, a2, ax, a = 0;	// Knightmare- start as 0
	vec3_t	minv, maxv;
	vec3_t	point;
	int		i;
	int		tag;
	int		ecount[2];
	float	scale;
	vec3_t	n1, n2;
	vec3_t	e1norm, e2norm;
	float	e1point, e2point;
	struct {
		int		pnum;
		vec3_t	point;
		vec3_t	normal;
	} etemp[2][2][2], e1vert[2], e2vert[2];

	// if phong shading disabled, or face not set for it,
	// just return old normal
	if (nocurve || !IsCurve(f))
	{
		VectorCopy(oldnormal, normal);
		return;
	}

#if 0
	tag = texinfo[f->texinfo].value;
	ecount[0] = 0;
	ecount[1] = 0;
	for (i=0; i<f->numedges; i++)
	{
		int	e;
		int	p1, p2;

		e = dsurfedges[f->firstedge + i];

		if (e >= 0)
		{
			p1 = dedges[e].v[0];
			p2 = dedges[e].v[1];
		}
		else
		{
			p1 = dedges[-e].v[1];
			p2 = dedges[-e].v[0];
		}

		GetVertexNormal(p1, tag, normal);
		return;	
	}
#else
	// set which two axis to use
	switch (dplanes[f->planenum].type)
	{
	case PLANE_X:
	case PLANE_ANYX:
		a1 = 1; // y
		a2 = 2; // z
		break;
	case PLANE_Y:
	case PLANE_ANYY:
		a1 = 0; // x
		a2 = 2; // z
		break;
	default:	// Knightmare added
	case PLANE_Z:
	case PLANE_ANYZ:
		a1 = 0; // x
		a2 = 1; // y
		break;
	}

	// get face's bounding rectangle
	minv[a1] = 999999;
	minv[a2] = 999999;
	maxv[a1] = -999999;
	maxv[a2] = -999999;
	for (i=0; i<f->numedges; i++)
	{
		int e;
		int p;

		e = dsurfedges[f->firstedge + i];

		if (e >= 0)
			p = dedges[e].v[0];
		else
			p = dedges[-e].v[1];

		if (dvertexes[p].point[a1] > maxv[a1])
			maxv[a1] = dvertexes[p].point[a1];
		if (dvertexes[p].point[a1] < minv[a1])
			minv[a1] = dvertexes[p].point[a1];

		if (dvertexes[p].point[a2] > maxv[a2])
			maxv[a2] = dvertexes[p].point[a2];
		if (dvertexes[p].point[a2] < minv[a2])
			minv[a2] = dvertexes[p].point[a2];
	}

	// tweak points that are beyond both axis boundaries
	VectorCopy(pnt, point);
	if (point[a1] > maxv[a1])
	{
		if (point[a2] > maxv[a2])
		{
			point[a1] = maxv[a1];
			point[a2] = maxv[a2];
		}
		else if (point[a2] < minv[a2])
		{
			point[a1] = maxv[a1];
			point[a2] = minv[a2];
		}
	}
	else if (point[a1] < minv[a1])
	{
		if (point[a2] > maxv[a2])
		{
			point[a1] = minv[a1];
			point[a2] = maxv[a2];
		}
		else if (point[a2] < minv[a2])
		{
			point[a1] = minv[a1];
			point[a2] = minv[a2];
		}
	}

	// find 2 edges to use for shading
	tag = texinfo[f->texinfo].value;
	ecount[0] = 0;
	ecount[1] = 0;
	for (i=0; i<f->numedges; i++)
	{
		int	e;
		int	p1, p2;

		e = dsurfedges[f->firstedge + i];

		if (e >= 0)
		{
			p1 = dedges[e].v[0];
			p2 = dedges[e].v[1];
		}
		else
		{
			p1 = dedges[-e].v[1];
			p2 = dedges[-e].v[0];
		}

		// if point is on a vertex, return apropriate normal
		if ((dvertexes[p1].point[0] == point[0]) &&
			(dvertexes[p1].point[1] == point[1]) &&
			(dvertexes[p1].point[2] == point[2]))
		{
			GetVertexNormal(p1, tag, normal);
			return;
		}

		// need to check along both axis for edges
		for (a=0; a<2; a++)
		{
			if (a == 0)
				ax = a1;
			else
				ax = a2;

			// if edge lies on this axis, move on
			if ((dvertexes[p1].point[ax] == point[ax]) &&
				(dvertexes[p2].point[ax] == point[ax]))
				continue;

			// if this edge is useable, save vertex data
			if (((dvertexes[p1].point[ax] >= point[ax]) &&
				 (dvertexes[p2].point[ax] <= point[ax])) ||
				((dvertexes[p1].point[ax] <= point[ax]) &&
				 (dvertexes[p2].point[ax] >= point[ax])))
			{
				etemp[a][ecount[a]][0].pnum = p1;
				etemp[a][ecount[a]][1].pnum = p2;
				VectorCopy(dvertexes[p1].point, etemp[a][ecount[a]][0].point);
				VectorCopy(dvertexes[p2].point, etemp[a][ecount[a]][1].point);

				ecount[a]++;
			}

			// found edges, exit loop
			if (ecount[a] == 2)
				break;
		}

		// if loop ended early, edges found, exit this loop too
		if (a != 2)
			break;
	}

	// this shouldn't happen anymore
	if (a == 2)
	{
		VectorCopy(oldnormal, normal);
		return;
	}

	// store vertex data for found edges
	VectorCopy(etemp[a][0][0].point, e1vert[0].point);
	VectorCopy(etemp[a][0][1].point, e1vert[1].point);
	GetVertexNormal(etemp[a][0][0].pnum, tag, e1vert[0].normal);
	GetVertexNormal(etemp[a][0][1].pnum, tag, e1vert[1].normal);

	VectorCopy(etemp[a][1][0].point, e2vert[0].point);
	VectorCopy(etemp[a][1][1].point, e2vert[1].point);
	GetVertexNormal(etemp[a][1][0].pnum, tag, e2vert[0].normal);
	GetVertexNormal(etemp[a][1][1].pnum, tag, e2vert[1].normal);

	// if edges found on second axis, switch axis for below equations
	if (a == 1)
	{
		int atemp;

		atemp = a1;
		a1 = a2;
		a2 = atemp;
	}

	// interpolate edge normals and get edge coords
	if (e1vert[0].point[a1] == e1vert[1].point[a1])
		scale = (e1vert[0].point[a2] - point[a2]) /
				(e1vert[0].point[a2] - e1vert[1].point[a2]);
	else
		scale = (e1vert[0].point[a1] - point[a1]) /
				(e1vert[0].point[a1] - e1vert[1].point[a1]);
	VectorScale(e1vert[0].normal, (1.0-scale), n1);
	VectorScale(e1vert[1].normal, scale, n2);
	VectorAdd(n1, n2, e1norm);
	VectorNormalize(e1norm, e1norm);
	e1point = (e1vert[0].point[a2] * (1.0-scale)) +
			  (e1vert[1].point[a2] * scale);

	if (e2vert[0].point[a1] == e2vert[1].point[a1])
		scale = (e2vert[0].point[a2] - point[a2]) /
				(e2vert[0].point[a2] - e2vert[1].point[a2]);
	else
		scale = (e2vert[0].point[a1] - point[a1]) /
				(e2vert[0].point[a1] - e2vert[1].point[a1]);
	VectorScale(e2vert[0].normal, (1.0-scale), n1);
	VectorScale(e2vert[1].normal, scale, n2);
	VectorAdd(n1, n2, e2norm);
	VectorNormalize(e2norm, e2norm);
	e2point = (e2vert[0].point[a2] * (1.0-scale)) +
			  (e2vert[1].point[a2] * scale);

	// interpolate final point normal
	if (e1point == e2point)
		scale = (e1point - point[a2]) / 0.000001;
	else
		scale = (e1point - point[a2]) / (e1point - e2point);
	VectorScale(e1norm, (1.0-scale), n1);
	VectorScale(e2norm, scale, n2);
	VectorAdd(n1, n2, normal);
	VectorNormalize(normal, normal);
	return;
#endif
}

float	sampleofs[5][2] =
{  {0,0}, {-0.25, -0.25}, {0.25, -0.25}, {0.25, 0.25}, {-0.25, 0.25} };

void BuildFacelights (int facenum)
{
	dface_t	*f;
	lightinfo_t	l[5];
	float		*styletable[MAX_LSTYLES];
	int			i, j;
	float		*spot;
	patch_t		*patch;
	int			numsamples;
	int			tablesize;
	facelight_t		*fl;
	// Argh!
	int		style;
	vec3_t	normal; // (2.00)
	float	*bouncelight; // stores only bouncing light (dk)

	f = &dfaces[facenum];

//	if ( texinfo[f->texinfo].flags & (SURF_SKY|SURF_FULLBRIGHT|SURF_FOGPLANE) ) // (dk)
//	if ( texinfo[f->texinfo].flags & (SURF_WARP|SURF_SKY|SURF_FOGPLANE) )
	if ( texinfo[f->texinfo].flags & (SURF_SKY|SURF_FOGPLANE) )	// Knightmare- allow lightmaps for warp surfaces
		return;	// non-lit texture

	memset (styletable,0, sizeof(styletable));

	if (extrasamples)
		numsamples = 5;
	else
		numsamples = 1;
	for (i=0 ; i<numsamples ; i++)
	{
		memset (&l[i], 0, sizeof(l[i]));
		l[i].surfnum = facenum;
		l[i].face = f;
		VectorCopy (dplanes[f->planenum].normal, l[i].facenormal);
		l[i].facedist = dplanes[f->planenum].dist;
		if (f->side)
		{
			VectorSubtract (vec3_origin, l[i].facenormal, l[i].facenormal);
			l[i].facedist = -l[i].facedist;
		}

		// get the origin offset for rotating bmodels
		VectorCopy (face_offset[facenum], l[i].modelorg);

		CalcFaceVectors (&l[i]);
		CalcFaceExtents (&l[i]);
		CalcPoints (&l[i], sampleofs[i][0], sampleofs[i][1]);
	}

	tablesize = l[0].numsurfpt * sizeof(vec3_t);
	styletable[0] = malloc(tablesize);
	memset (styletable[0], 0, tablesize);
	// (dk)
	bouncelight = malloc(tablesize);
	memset (bouncelight, 0, tablesize);

	fl = &facelight[facenum];
	fl->numsamples = l[0].numsurfpt;
	fl->origins = malloc (tablesize);
	memcpy (fl->origins, l[0].surfpt, tablesize);

	for (i=0 ; i<l[0].numsurfpt ; i++)
	{
		for (j=0 ; j<numsamples ; j++)
		{
			//Argh! - use phong function to get normal,
			//        and send to GatherSampleLight (2.00)
			GetPhongNormal(f, l[j].surfpt[i], l[0].facenormal, normal);

			// added bouncelight to args (dk)
			GatherSampleLight (l[j].surfpt[i], normal, styletable,
				bouncelight, i*3, tablesize, 1.0/numsamples);
		}

		// contribute the sample to one or more patches
		// only add bouncing light (dk)
		AddSampleToPatch (l[0].surfpt[i], bouncelight+i*3, facenum);
	}
	// done with this (dk)
	free (bouncelight);

	// average up the direct light on each patch for radiosity
	for (patch = face_patches[facenum] ; patch ; patch=patch->next)
	{
		if (patch->samples)
		{
			VectorScale (patch->samplelight, 1.0/patch->samples, patch->samplelight);
		}
	}

	// Argh! - allocate style table if needed
	style = face_patches[facenum]->style;
	if (!styletable[style])
	{
		styletable[style] = malloc (tablesize);
		memset (styletable[style], 0, tablesize);
	}

	for (i=0 ; i<MAX_LSTYLES ; i++)
	{
		if (!styletable[i])
			continue;
		if (fl->numstyles == MAX_STYLES)
			break;
		fl->samples[fl->numstyles] = styletable[i];
		fl->stylenums[fl->numstyles] = i;
		fl->numstyles++;
	}

	// Argh! - make surface lights bright, handles lightmin and
	//         non style-0 (replaces id code)
	if (IsLight(face_patches[facenum]->face))
	{
		vec3_t lightmin;

		// calc lightmin
		VectorNormalize(face_patches[facenum]->baselight, lightmin);
		VectorScale(lightmin, face_patches[facenum]->lightmin, lightmin);

		for (i=0 ; i<l[0].numsurfpt ; i++)
		{
			spot = styletable[style] + i*3;
			VectorAdd (spot, face_patches[facenum]->baselight, spot);

			// set lightmin if needed
		//	if (lightmin)
			if ( VectorLength(lightmin) > 0 )	// Knightmare- fix warning
				VectorAdd (spot, lightmin, spot);
		}
	}

}


// Argh! - returns true if faces are the same, or are touching
//         (share a vertex, looks better than edge test) (rev2.00)
qboolean FacesTouching(dface_t *f1, dface_t *f2)
{
	int i, j;
	int e1, e2;
	int p1, p2;

	// quick return if same face
	if (f1 == f2)
		return true;

	// go through all face1 edges
	for (i=0; i<f1->numedges; i++)
	{
		e1 = dsurfedges[f1->firstedge + i];

		// get vertex for edge
		if (e1 >= 0)
			p1 = dedges[e1].v[0];
		else
			p1 = dedges[-e1].v[1];

		// go through all face2 edges
		for (j=0; j<f2->numedges; j++)
		{
			e2 = dsurfedges[f2->firstedge + j];

			// get vertex for edge
			if (e2 >= 0)
				p2 = dedges[e2].v[0];
			else
				p2 = dedges[-e2].v[1];

			// if vertices same, faces touch
			if (p1 == p2)
				return true;
		}
	}

	return false;
}

/*
=============
FinalLightFace

Add the indirect lighting on top of the direct
lighting and save into final map format
=============
*/
void FinalLightFace (int facenum)
{
	dface_t		*f;
	int			i, j, k, st;
	vec3_t		lb;
	patch_t		*patch;
	triangulation_t	*trian = NULL;	// Knightmare- start as null
	facelight_t	*fl;
	float		minlight = 0;
	float		max, newmax;
	byte		*dest;
	int			pfacenum;
	vec3_t		facemins, facemaxs;
	// Argh!
	float		d1, d2;   // for distance check (2.00)
	char		*name;    // entity name
	vec3_t		eambient; // entity-specific ambient term
	char		*amb;     // read in eambient

	f = &dfaces[facenum];
	fl = &facelight[facenum];

//	if ( texinfo[f->texinfo].flags & (SURF_SKY|SURF_FULLBRIGHT|SURF_FOGPLANE) ) // (dk)
//	if ( texinfo[f->texinfo].flags & (SURF_WARP|SURF_SKY|SURF_FOGPLANE) )
	if ( texinfo[f->texinfo].flags & (SURF_SKY|SURF_FOGPLANE) )// Knightmare- allow lightmaps for warp surfaces
		return;	// non-lit texture

	ThreadLock ();
	f->lightofs = lightdatasize;
	lightdatasize += fl->numstyles*(fl->numsamples*3);

	if (lightdatasize > MAX_MAP_LIGHTING)
		Error ("MAX_MAP_LIGHTING");
	ThreadUnlock ();

	f->styles[0] = 0;
	f->styles[1] = f->styles[2] = f->styles[3] = 0xff;

	//
	// set up the triangulation
	//
	if (numbounce > 0)
	{
		ClearBounds (facemins, facemaxs);
		for (i=0 ; i<f->numedges ; i++)
		{
			int		ednum;

			ednum = dsurfedges[f->firstedge+i];
			if (ednum >= 0)
				AddPointToBounds (dvertexes[dedges[ednum].v[0]].point,
				facemins, facemaxs);
			else
				AddPointToBounds (dvertexes[dedges[-ednum].v[1]].point,
				facemins, facemaxs);
		}

		trian = AllocTriangulation (&dplanes[f->planenum]);

		/*
		// for all faces on the plane, add the nearby patches
		// to the triangulation
		// Argh! - end when pfacenum == -1, fixes qrad3 bug (2.00)
		for (pfacenum = planelinks[f->side][f->planenum]
			; pfacenum != -1 ; pfacenum = facelinks[pfacenum])
		*/

		// Argh! - need to check all faces now because of curves (2.00)
		for (pfacenum = 0; pfacenum < numfaces; pfacenum++)
		{
			dface_t *f2; // Argh! - (2.00)

			f2 = &dfaces[pfacenum];

			// Argh! - different side, move on (2.00)
			if (f->side != f2->side)
				continue;

			// Argh! - check for same plane (2.00)
			if (f->planenum == f2->planenum)
			{
				// only consider faces that are the same or
				// touching for triangulation, checking can
				// be disabled on the command line (1.50)
				if (stopbleed && !FacesTouching(f, f2))
					continue;
			}
			else
			{
				// if phong shading disabled, or not a curve
				// face, or faces not the same curve, move on
				if (nocurve || !IsCurve(f) ||
					(texinfo[f->texinfo].value != texinfo[f2->texinfo].value))
					continue;

				// curve faces must always touch to be considered
				if (!FacesTouching(f, f2))
					continue;
			}

			for (patch = face_patches[pfacenum] ; patch ; patch=patch->next)
			{
				// distance check
				for (i=0 ; i < 3 ; i++)
				{
					// Argh! - define here to simplify/speed up code (rev2.00)
					d1 = (facemins[i] - patch->origin[i])/2;
					d2 = (patch->origin[i] - facemaxs[i])/2;

					// Argh! - handle special patch subdivision
					if (IsLight(patch->face))
					{
						if (IsSky(patch->face))
						{
							if ((d1 > subdivsky) || (d2 > subdivsky))
								break;
						}
						else if (IsWarp(patch->face))
						{
							if ((d1 > subdivwarp) || (d2 > subdivwarp))
								break;
						}
						else
						{
							if ((d1 > subdivlight) || (d2 > subdivlight))
								break;
						}
					}
					else
					{
						if (IsCurve(patch->face)) // (2.00)
						{
							if ((d1 > subdivcurve) || (d2 > subdivcurve))
								break;
						}
						else
						{
							if ((d1 > subdiv) || (d2 > subdiv))
								break;
						}
					}
				}
				if (i != 3)
					continue;	// not needed for this face

				AddPointToTriangulation (patch, trian);
			}
		}

		for (i=0 ; i<trian->numpoints ; i++)
			memset (trian->edgematrix[i], 0, trian->numpoints*sizeof(trian->edgematrix[0][0]));
		TriangulatePoints (trian);
	}

	//
	// sample the triangulation
	//

	// Argh! - eambient & minlight don't apply to worldspawn (fixed1.02)
	VectorClear(eambient);
	name = ValueForKey (face_entity[facenum], "classname");
	if (strcmp(name, "worldspawn"))
	{
		// Argh! - entity-specific ambient term
		amb = ValueForKey(face_entity[facenum], "_ambient");
		if (amb[0])
		{
			int n = sscanf(amb, "%f %f %f", &eambient[0], &eambient[1], &eambient[2]);
			if (n < 3)
			{
				eambient[1] = eambient[0];
				eambient[2] = eambient[0];
			}
		}

		// _minlight allows models that have faces that would not be
		// illuminated to receive a mottled light pattern instead of
		// black
		minlight = FloatForKey(face_entity[facenum], "_minlighta");
		if (!minlight)
			minlight = FloatForKey(face_entity[facenum], "_minlight") * 128;
	}

	dest = &dlightdata[f->lightofs];

	if (fl->numstyles > MAXLIGHTMAPS)
	{
		fl->numstyles = MAXLIGHTMAPS;
		printf ("face with too many lightstyles: (%f %f %f)\n",
			face_patches[facenum]->origin[0],
			face_patches[facenum]->origin[1],
			face_patches[facenum]->origin[2]
			);
	}

	for (st=0 ; st<fl->numstyles ; st++)
	{
		f->styles[st] = fl->stylenums[st];
		for (j=0 ; j<fl->numsamples ; j++)
		{
			VectorCopy ( (fl->samples[st]+j*3), lb);

			// Argh! - only work on style-0, fixes qrad3 bug (1.03)
			if (st == 0)
			{
				if (numbounce > 0)
				{
					vec3_t add;

					// Argh! - SampleTriangulation needs to know if curve (2.00)
					if (trian != NULL)	// Knightmare- check pointer first!
						SampleTriangulation (fl->origins + j*3, trian, IsCurve(f), add);

					// Argh! - only save bounced light if desired (2.00)
					if (onlybounce)
					{
						VectorCopy(add, lb);
					}
					else
					{
						VectorAdd (lb, add, lb);
					}
				}

				// add an ambient term if desired
				lb[0] += ambient[0];
				lb[1] += ambient[1];
				lb[2] += ambient[2];

				// Argh! - add entity-specific ambient term
				lb[0] += eambient[0];
				lb[1] += eambient[1];
				lb[2] += eambient[2];

				// Argh! - account for minlimit
				if (minlimit)
				{
					if (lb[0] < ambient[0])
						lb[0] = ambient[0];
					if (lb[1] < ambient[1])
						lb[1] = ambient[1];
					if (lb[2] < ambient[2])
						lb[2] = ambient[2];
				}
			}
			else // non lightstyle 0
			{
				minlight = 0;
			}

			// Argh! - nocolor conversions
			if (nocolor == 1)
			{
				double value = lb[0] * 0.299 + lb[1] * 0.588 + lb[2] * 0.113;

				lb[0] = value;
				lb[1] = value;
				lb[2] = value;
			}
			else if (nocolor == 2)
			{
				double value = (lb[0] + lb[1] + lb[2]) / 3.0;

				lb[0] = value;
				lb[1] = value;
				lb[2] = value;
			}

			VectorScale (lb, lightscale, lb);

			// we need to clamp without allowing hue to change
			for (k=0 ; k<3 ; k++)
				if (lb[k] < 1)
					lb[k] = 1;
			max = lb[0];
			if (lb[1] > max)
				max = lb[1];
			if (lb[2] > max)
				max = lb[2];
			newmax = max;
			if (newmax < 0)
				newmax = 0;		// roundoff problems
			if (newmax < minlight)
			{
				newmax = minlight + (rand()%48);
			}
			if (newmax > maxlight)
				newmax = maxlight;

			for (k=0 ; k<3 ; k++)
			{
				*dest++ = lb[k]*newmax/max;
			}
		}
	}

	if (numbounce > 0)
		FreeTriangulation (trian);
}
