// patches.c

// ArghRad source - modified version of qrad3 source by id Software.
// ArghRad code modifications may not be used without permission
// from author Tim Wright. (Argh!)

// This version modified for Daikatana (dk)

#include "qrad.h"

vec3_t	texture_reflectivity[MAX_MAP_TEXINFO];

/*
===================================================================

  TEXTURE LIGHT VALUES

===================================================================
*/


// Argh! - moved pak-loading code to pak.c & pak.h (2.00)
#if 0
// moved from qrad3.c, now returns pal dir name (dk)
char *SetPaletteDirectory (void)
{
	int			i;
	entity_t	*e;
	char		*classname, *palette;

	for (i=0 ; i<num_entities ; i++)
	{
		e = &entities[i];
		classname = ValueForKey (e, "classname");
		if (!strcmp (classname, "worldspawn"))
		{
			palette = ValueForKey (e, "palette");
			if (!palette)
				palette = ValueForKey (e, "palettedir");

			if (palette)
				return palette;
			else
				return "e1m6";
		}
	}

	return "";
}
#endif

void CalcTextureReflectivity (void)
{
	int				i;
	int				j, k, texels;
	int				color[3];
	int				texel;
	byte			*palette;
	char			path[1024];
	float			r, scale;
	miptex_t		*mt;
	// Argh! - (1.50)
	int				size;	// for LoadPakFile
	byte			*data;	// for LoadPakFile
//	char			*palettedir; // (dk)

	// Argh! - load all pakdirs into memory
	LoadPakdirs();

	// load the palette
	// Argh! - unfortunately it's too difficult to use LoadPakFile
	//         to get a colormap.bmp, so they must be extracted (dk)
	//         (will look into it better in the future)
//	if (!(palettedir = SetPaletteDirectory()))
//		Error ("no \"palette\" found in worldspawn\n");

	// Knightmare 7/12/12- support loading palette from paks
//	sprintf (path, "%spics/colormap.pcx", gamedir);
//	Load256Image (path, NULL, &palette, NULL, NULL);
	sprintf (path, "pics/colormap.pcx");
	Load256ImageNew (path, false, NULL, &palette, NULL, NULL);
	if (!palette)
		Error ("Couldn't load %s\n", path);
	// end Knightmare

	// allways set index 0 even if no textures
	texture_reflectivity[0][0] = 0.5;
	texture_reflectivity[0][1] = 0.5;
	texture_reflectivity[0][2] = 0.5;

	for (i=0 ; i<numtexinfo ; i++)
	{
		// see if an earlier texinfo allready got the value
		for (j=0 ; j<i ; j++)
		{
			if (!strcmp (texinfo[i].texture, texinfo[j].texture))
			// extsurfinfo (dk)
		/*	if (!strcmp (texinfo[i].texture, texinfo[j].texture) &&
				extsurfinfo[i].color [0] <= 0.0 &&
				extsurfinfo[i].color [1] <= 0.0 &&
				extsurfinfo[i].color [2] <= 0.0)	*/
			{
				VectorCopy (texture_reflectivity[j], texture_reflectivity[i]);
				break;
			}
		}
		if (j != i)
			continue;

		// load the wal file
		// Argh! - use new pak load function (rev1.50)
		sprintf(path, "textures/%s.wal", texinfo[i].texture);
		size = LoadPakFile(path, &data);
		mt = (miptex_t *)data;
		if (!mt)
		{
			printf ("Couldn't load %s\n", path);
			texture_reflectivity[i][0] = 0.5;
			texture_reflectivity[i][1] = 0.5;
			texture_reflectivity[i][2] = 0.5;
			continue;
		}

	//	if (mt->version != MIPTEX_VERSION) // (dk)
	//		Error ("Miptex has wrong version. Regrab your .wal files\n");

		texels = LittleLong(mt->width)*LittleLong(mt->height);
		color[0] = color[1] = color[2] = 0;

		// process new wal format (dk)
	/*	if ((extsurfinfo[i].color[0] > 0) || (extsurfinfo[i].color[1] > 0) || (extsurfinfo[i].color[2] > 0))
		{
			for (j=0 ; j<texels ; j++)
			{
				for (k=0 ; k<3 ; k++)
					color[k] += extsurfinfo[i].color[k];
			}
		}
		else	*/
		{
			for (j=0 ; j<texels ; j++)
			{
				texel = ((byte *)mt)[LittleLong(mt->offsets[0]) + j];
				// make sure the texel is not the transparent color
				if( texel != 255 )
				{
					for (k=0 ; k<3 ; k++)
						color[k] += palette[texel*3+k];
				}
			}
		}

		for (j=0 ; j<3 ; j++)
		{
		//	if (!(r = color[j]/texels/255.0))
			r = color[j]/texels/255.0;	// Knightmare- warning fix
			if (!r)
				r = 0.1; // hack for black textures (all textures have to have SOME reflectivity) (dk)
			texture_reflectivity[i][j] = r;
		}

		// scale the reflectivity up, because the textures are
		// so dim
		scale = ColorNormalize (texture_reflectivity[i],
			texture_reflectivity[i]);
		if (scale < 0.5)
		{
			scale *= 2;
			VectorScale (texture_reflectivity[i], scale, texture_reflectivity[i]);
		}

		// Argh! - free texture data (rev1.50)
		free(data);
	}

    // Argh! - free palette and pakdir memory
	free(palette);
    FreePakdirs();
}

/*
=======================================================================

MAKE FACES

=======================================================================
*/

winding_t	*WindingFromFace (dface_t *f)
{
	int			i;
	int			se;
	dvertex_t	*dv;
	int			v;
	winding_t	*w;

	w = AllocWinding (f->numedges);
	w->numpoints = f->numedges;

	for (i=0 ; i<f->numedges ; i++)
	{
		se = dsurfedges[f->firstedge + i];
		if (se < 0)
			v = dedges[-se].v[1];
		else
			v = dedges[se].v[0];

		dv = &dvertexes[v];
		VectorCopy (dv->point, w->p[i]);
	}

	RemoveColinearPoints (w);

	return w;
}

void BaseLightForFace (dface_t *f, vec3_t color)
{
	texinfo_t	*tx;

	//
	// check for light emited by texture
	//
	tx = &texinfo[f->texinfo];
	if (!(tx->flags & SURF_LIGHT) || tx->value == 0)
	{
		VectorClear (color);
		return;
	}

	VectorScale (texture_reflectivity[f->texinfo], tx->value, color);
}

float	totalarea;

void MakePatchForFace (int fn, winding_t *w, entity_t *ent)
// Argh! - added ent(ity) to function params
{
	dface_t *f;
	float	area;
	patch_t		*patch;
	dplane_t	*pl;
	int			i;
	vec3_t		color;
	dleaf_t		*leaf;

	f = &dfaces[fn];

	area = WindingArea (w);
	totalarea += area;

	patch = &patches[num_patches];
	if (num_patches == MAX_PATCHES)
		Error ("num_patches == MAX_PATCHES");
	patch->next = face_patches[fn];
	face_patches[fn] = patch;

	patch->face = f; // Argh! - pointer to patch's face (2.00)

	patch->winding = w;

	if (f->side)
		patch->plane = &backplanes[f->planenum];
	else
		patch->plane = &dplanes[f->planenum];
	if (face_offset[fn][0] || face_offset[fn][1] || face_offset[fn][2] )
	{	// origin offset faces must create new planes
		if (numplanes + fakeplanes >= MAX_MAP_PLANES)
			Error ("numplanes + fakeplanes >= MAX_MAP_PLANES");
		pl = &dplanes[numplanes + fakeplanes];
		fakeplanes++;

		*pl = *(patch->plane);
		pl->dist += DotProduct (face_offset[fn], pl->normal);
		patch->plane = pl;
	}

	WindingCenter (w, patch->origin);
	VectorAdd (patch->origin, patch->plane->normal, patch->origin);
	leaf = PointInLeaf(patch->origin);
	patch->cluster = leaf->cluster;
	if (patch->cluster == -1)
		qprintf ("patch->cluster == -1\n");

	// Argh! - set patch normal (2.00)
	GetPhongNormal(patch->face, patch->origin, patch->plane->normal, patch->normal);

	patch->area = area;
	if (patch->area < 1) // Argh! - made <= into < (1.50)
		patch->area = 1;

	VectorCopy (texture_reflectivity[f->texinfo], patch->reflectivity);

	// non-bmodel patches can emit light
	// Argh! - let bmodels emit light if desired
	if (bmodlight || fn < dmodels[0].numfaces)
	{
		BaseLightForFace (f, patch->baselight);

		// Argh! - if bmodel, get all options
		if ( fn >= dmodels[0].numfaces )
		{
			char *_color;
			char *focus;
			float light;
			float lightmin;
			int style;

			// get color
			_color = ValueForKey(ent, "_color");
			if (_color && _color[1])
			{
				float scale;

				sscanf (_color, "%f %f %f", &color[0], &color[1], &color[2]);
				ColorNormalize (color, color);

				// tweak reflectivity to new color
				scale = max(fabs(patch->reflectivity[0]), max(fabs(patch->reflectivity[1]), fabs(patch->reflectivity[2])));
				VectorScale(color, scale, patch->reflectivity);
			}
			else
			{
				ColorNormalize (patch->reflectivity, color);
			}

			// get focus
			focus = ValueForKey (ent, "_focus");
			if (focus[0])
				patch->focus = atof(focus);
			else
				patch->focus = 1.0;
			if (patch->focus < 0.0)
				patch->focus = 0.0;

			// get style
			style = FloatForKey (ent, "style");
			if (!style)
				style = FloatForKey (ent, "_style");
			patch->style = style;

			// get light
			light = FloatForKey (ent, "light");
			if (!light)
				light = FloatForKey (ent, "_light");
			if (texinfo[f->texinfo].flags & SURF_LIGHT)
			{
				// tweak baselight to new reflectivity
				if (light)
				{
					VectorScale(patch->reflectivity, light, patch->baselight);
				}
				else
				{
					VectorScale(patch->reflectivity, texinfo[f->texinfo].value, patch->baselight);
				}
			}

			// get min facelight brightness
			lightmin = FloatForKey(ent, "_lightmin");
			patch->lightmin = lightmin;
		}
		else // not bmodel, init fields
		{
			ColorNormalize (patch->reflectivity, color);
			patch->focus = 1.0;
			patch->style = 0.0;
			patch->lightmin = 0.0;
		}

		// Argh! - fabs() accounts for antilights
		for (i=0 ; i<3 ; i++)
			patch->baselight[i] *= fabs(color[i]);

		VectorCopy (patch->baselight, patch->totallight);
	}
	num_patches++;
}


entity_t *EntityForModel (int modnum)
{
	int		i;
	char	*s;
	char	name[16];

	sprintf (name, "*%i", modnum);
	// search the entities for one using modnum
	for (i=0 ; i<num_entities ; i++)
	{
		s = ValueForKey (&entities[i], "model");
		if (!strcmp (s, name))
			return &entities[i];
	}

	return &entities[0];
}

void MakePatches (void)
{
	int		i, j, k;
	dface_t	*f;
	int		fn;
	winding_t	*w;
	dmodel_t	*mod;
	vec3_t		origin;
	entity_t	*ent;

	qprintf ("%i faces\n", numfaces);

	for (i=0 ; i<nummodels ; i++)
	{
		mod = &dmodels[i];
		ent = EntityForModel (i);
		// bmodels with origin brushes need to be offset into their
		// in-use position
		GetVectorForKey (ent, "origin", origin);

		for (j=0 ; j<mod->numfaces ; j++)
		{
			fn = mod->firstface + j;
			face_entity[fn] = ent;
			VectorCopy (origin, face_offset[fn]);
			f = &dfaces[fn];
			w = WindingFromFace (f);
			for (k=0 ; k<w->numpoints ; k++)
			{
				VectorAdd (w->p[k], origin, w->p[k]);
			}
			MakePatchForFace (fn, w, ent); // Argh! - send ent too
		}
	}

	qprintf ("%i square feet\n", (int)(totalarea/64));
}

/*
=======================================================================

SUBDIVIDE

=======================================================================
*/

void FinishSplit (patch_t *patch, patch_t *newp)
{
	dleaf_t		*leaf;

	VectorCopy (patch->baselight, newp->baselight);
	VectorCopy (patch->totallight, newp->totallight);
	VectorCopy (patch->reflectivity, newp->reflectivity);
	newp->plane = patch->plane;
	// Argh!
	newp->focus = patch->focus;
	newp->style = patch->style;
	newp->lightmin = patch->lightmin;
	newp->face = patch->face; // (2.00)

	patch->area = WindingArea (patch->winding);
	newp->area = WindingArea (newp->winding);

	// Argh! - made <= into < (2.00)
	if (patch->area < 1)
		patch->area = 1;
	if (newp->area < 1)
		newp->area = 1;

	WindingCenter (patch->winding, patch->origin);
	VectorAdd (patch->origin, patch->plane->normal, patch->origin);
	leaf = PointInLeaf(patch->origin);
	patch->cluster = leaf->cluster;
	if (patch->cluster == -1)
		qprintf ("patch->cluster == -1\n");

	// Argh! - set patch normal (2.00)
	GetPhongNormal(patch->face, patch->origin, patch->plane->normal, patch->normal);

	WindingCenter (newp->winding, newp->origin);
	VectorAdd (newp->origin, newp->plane->normal, newp->origin);
	leaf = PointInLeaf(newp->origin);
	newp->cluster = leaf->cluster;
	if (newp->cluster == -1)
		qprintf ("patch->cluster == -1\n");

	// Argh! - set patch normal (2.00)
	GetPhongNormal(newp->face, newp->origin, newp->plane->normal, newp->normal);
}


/*
=============
DicePatch

Chops the patch by a global grid
=============
*/
void	DicePatch (patch_t *patch)
{
	winding_t *w, *o1, *o2;
	vec3_t	mins, maxs;
	vec3_t	split;
	vec_t	dist;
	int		i;
	patch_t	*newp;

	w = patch->winding;
	WindingBounds (w, mins, maxs);
	for (i=0 ; i<3 ; i++)
	{
		// Argh! - handle special patch subdivision
		if (IsLight(patch->face))
		{
			if (IsSky(patch->face))
			{
				if (floor((mins[i]+1)/subdivsky) < floor((maxs[i]-1)/subdivsky))
					break;
			}
			else if (IsWarp(patch->face))
			{
				if (floor((mins[i]+1)/subdivwarp) < floor((maxs[i]-1)/subdivwarp))
					break;
			}
			else
			{
				if (floor((mins[i]+1)/subdivlight) < floor((maxs[i]-1)/subdivlight))
					break;
			}
		}
		else
		{
			if (IsCurve(patch->face)) // (2.00)
			{
				if (floor((mins[i]+1)/subdivcurve) < floor((maxs[i]-1)/subdivcurve))
					break;
			}
			else
			{
				if (floor((mins[i]+1)/subdiv) < floor((maxs[i]-1)/subdiv))
					break;
			}
		}
	}
	if (i == 3)
	{
		// no splitting needed
		return;
	}

	//
	// split the winding
	//
	VectorCopy (vec3_origin, split);
	split[i] = 1;
	// Argh! - handle special patch subdivision
	if (IsLight(patch->face))
	{
		if (IsSky(patch->face))
			dist = subdivsky*(1+floor((mins[i]+1)/subdivsky));
		else if (IsWarp(patch->face))
			dist = subdivwarp*(1+floor((mins[i]+1)/subdivwarp));
		else
			dist = subdivlight*(1+floor((mins[i]+1)/subdivlight));
	}
	else
	{
		if (IsCurve(patch->face)) // (2.00)
			dist = subdivcurve*(1+floor((mins[i]+1)/subdivcurve));
		else
			dist = subdiv*(1+floor((mins[i]+1)/subdiv));
	}
	ClipWindingEpsilon (w, split, dist, ON_EPSILON, &o1, &o2);

	//
	// create a new patch
	//
	if (num_patches == MAX_PATCHES)
		Error ("MAX_PATCHES");
	newp = &patches[num_patches];
	num_patches++;

	newp->next = patch->next;
	patch->next = newp;

	patch->winding = o1;
	newp->winding = o2;

	FinishSplit (patch, newp);

	DicePatch (patch);
	DicePatch (newp);
}

void SubdividePatches (void)
{
	int		i, num;

	if (subdiv < 1)
		return;

	num = num_patches;	// because the list will grow
	for (i=0 ; i<num ; i++)
	{
		DicePatch (&patches[i]);
	}
	qprintf ("%i patches after subdivision\n", num_patches);
}

