// qrad.h

// ArghRad source - modified version of qrad3 source by id Software.
// ArghRad code modifications may not be used without permission
// from author Tim Wright. (Argh!)

// This version modified for Daikatana (dk)

#include "../common/cmdlib.h"
#include "../common/mathlib.h"
#include "../common/bspfile.h"
#include "../common/polylib.h"
#include "../common/threads.h"
#include "../common/lbmlib.h"
#include "../common/pak.h" // (2.00)

#ifdef WIN32
#include <windows.h>
#endif

// Argh! - commented out unused declarations (2.00)


// Argh! - revised macros to take the face_t and just perform
//         the bitwise-and, added new curve flag and macro (rev2.00)
#define	IsLight(f)	(texinfo[f->texinfo].flags & SURF_LIGHT)
#define	IsSky(f)	(texinfo[f->texinfo].flags & SURF_SKY)
#define	IsWarp(f)	(texinfo[f->texinfo].flags & SURF_WARP)
#define	IsCurve(f)	(texinfo[f->texinfo].flags & SURF_CURVE)


typedef enum
{
	emit_surface,
	emit_point,
	emit_spotlight,
	emit_sunlight	// Argh! - sunlight
} emittype_t;


typedef struct directlight_s
{
	struct directlight_s *next;
	emittype_t	type;

	float	intensity;
	int		style;
	vec3_t	origin;
	vec3_t	color;
	vec3_t	normal;		// for surfaces and spotlights
	float	stopdot;	// for spotlights
	// Argh!
	float		fade;	// for pointlights, dist fade
	float		afade;	// for pointlights, angle fade
	float		focus;	// for surf/spotlights, focus fade
	int			falloff;// falloff type: 0=linear, 1=inverse, 2=invsquare
	dface_t		*face;	// emit_surface face (2.00)
	qboolean	bounce; // true if light bounces (dk)
	float		cap;    // max brightness cap (dk)
} directlight_t;


// the sum of all tranfer->transfer values for a given patch
// should equal exactly 0x10000, showing that all radiance
// reaches other patches
typedef struct
{
	unsigned short	patch;
	unsigned short	transfer;
} transfer_t;


#define	MAX_PATCHES	65000			// larger will cause 32 bit overflows

#define	MAX_LSTYLES	256

typedef struct patch_s
{
	winding_t	*winding;
	struct patch_s		*next;		// next in face
	int			numtransfers;
	transfer_t	*transfers;

	int			cluster;			// for pvs checking
	vec3_t		origin;
	dplane_t	*plane;

	vec3_t		totallight;			// accumulated by radiosity
									// does NOT include light
									// accounted for by direct lighting
	float		area;

	// illuminance * reflectivity = radiosity
	vec3_t		reflectivity;
	vec3_t		baselight;			// emissivity only

	// each style 0 lightmap sample in the patch will be
	// added up to get the average illuminance of the entire patch
	vec3_t		samplelight;
	int			samples;			// for averaging direct light

	// Argh! (reminder: when adding fields, edit FinishSplit())
	float		focus;		// directional focus falloff
	float		lightmin;	// min brightness of face light
	int			style;		// style of face light
	vec3_t		normal;		// plane normal (2.00)
	dface_t		*face;		// pointer to face, replaces flags (2.00)
} patch_t;


// Argh! - vertex normal data for phong shading (2.00)
typedef struct vnorms_s
{
	int				tag;
//	double			doubleNormal[3];
	vec3_t			normal;
	struct vnorms_s	*next;
} vnorms_t;

vnorms_t	*vertnorms[MAX_MAP_VERTS];


extern	patch_t		*face_patches[MAX_MAP_FACES];
extern	entity_t	*face_entity[MAX_MAP_FACES];
extern	vec3_t		face_offset[MAX_MAP_FACES];		// for rotating bmodels
extern	patch_t		patches[MAX_PATCHES];
extern	unsigned	num_patches;

extern	int		leafparents[MAX_MAP_LEAFS];
extern	int		nodeparents[MAX_MAP_NODES];

extern	float	lightscale;


void MakeShadowSplits (void);

//==============================================


void BuildVisMatrix (void);
qboolean CheckVisBit (unsigned p1, unsigned p2);

//==============================================

void LinkPlaneFaces (void);

extern	qboolean	extrasamples;
extern int numbounce;

extern	directlight_t	*directlights[MAX_MAP_LEAFS];

extern	byte	nodehit[MAX_MAP_NODES];

void BuildLightmaps (void);

void BuildFacelights (int facenum);

// Argh - (2.00)
void GetPhongNormal (dface_t *face, vec3_t pnt, vec3_t oldnormal, vec3_t normal);

void FinalLightFace (int facenum);

qboolean PvsForOrigin (vec3_t org, byte *pvs);

int TestLine_r (int node, vec3_t start, vec3_t stop);

void CreateDirectLights (void);

dleaf_t		*PointInLeaf (vec3_t point);


extern	dplane_t	backplanes[MAX_MAP_PLANES];
extern	int			fakeplanes;					// created planes for origin offset


extern	float	subdiv;
// Argh! - special patch subdivision
extern	float	subdivsky;
extern	float	subdivwarp;
extern	float	subdivlight;
extern	float	subdivcurve; // (2.00)

// Argh! - global light settings
extern	vec3_t		ambient;
extern	float		maxlight;
extern	qboolean	bmodlight;
extern	qboolean	nocurve;	// (2.00)

extern	float	direct_scale;
extern	float	entity_scale;

int	PointInLeafnum (vec3_t point);
void MakeTnodes (dmodel_t *bm);
void MakePatches (void);
void SubdividePatches (void);
void PairEdges (void);
void CalcTextureReflectivity (void);

// Argh! - from SiN qrad3 util (1.50)
void FreeDirectLights ( void );
