// qrad.c

// ArghRad source - modified version of qrad3 source by id Software.
// ArghRad code modifications may not be used without permission
// from author Tim Wright. (Argh!)

// notes:
// All ArghRad modifications have an "Argh!" comment for
// identification, those without a version number refer to the
// initial 1.00 release.  "rev" indicates an earlier mod was
// revised in the specified version, and "fixed" indicates an
// earlier mod was fixed in the specified version.
// The ArghRad Win32 executable is compiled using MSVC++4.0,
// configured for Pentium processor and max speed optimization.

// This version modified for Daikatana, mods indicated by "dk"

#include "qrad.h"

/*
NOTES
-----
every surface must be divided into at least two patches each axis
*/

// Argh! - reorganized old vars (rev2.00)

void BuildLightmaps (void);
int TestLine (vec3_t start, vec3_t stop);

patch_t		*face_patches[MAX_MAP_FACES];
entity_t	*face_entity[MAX_MAP_FACES];
patch_t		patches[MAX_PATCHES];
unsigned	num_patches;

vec3_t		radiosity[MAX_PATCHES];		// light leaving a patch
vec3_t		illumination[MAX_PATCHES];	// light arriving at a patch

vec3_t		face_offset[MAX_MAP_FACES];	// for rotating bmodels
dplane_t	backplanes[MAX_MAP_PLANES];
int			fakeplanes;					// created planes for origin offset

char		inbase[32], outbase[32];
char		source[1024];

int		junk;

qboolean	nopvs;
qboolean	dumppatches;
qboolean	glview;

int		numbounce = 8;
qboolean	extrasamples;

float	maxlight = 196;

float	lightscale = 1.0;
float	direct_scale =	0.4;
float	entity_scale =	1.0;

float	subdiv = 64;

// Argh! - new/changed global vars
vec3_t	ambient;

qboolean	stopbleed = true;	// true prevents light bleeds (1.50)
qboolean	bmodlight = true;	// true makes bmodels emit light
qboolean	minlimit = true;	// true keeps antilights above minlight
qboolean	onlybounce = false;	// true only saves bounced light (2.00)
qboolean	nocurve = false;	// true disables phong shading (2.00)
qboolean	splotchfix = true;	// true fixes plain surf_lights (2.01)

int	nocolor;	// 0=color, 1=nocolor, 2=nocoloru

float	subdivsky = 0;	// subdiv for sky surface light
float	subdivwarp = 0;	// subdiv for warp surface light
float	subdivlight = 0;// subdiv for other surface light
float	subdivcurve = 0;// subdiv for phong shaded surfaces (2.00)

float	radmin = 0.0;	// min cutoff for radiosity transfers (1.50)

vnorms_t	*vertnorms[MAX_MAP_VERTS];	// phong normals at vertex (2.00)

/*
===================================================================

MISC

===================================================================
*/

void MakeBackplanes (void)
{
	int		i;

	for (i=0 ; i<numplanes ; i++)
	{
		backplanes[i].dist = -dplanes[i].dist;
		VectorSubtract (vec3_origin, dplanes[i].normal, backplanes[i].normal);
	}
}

int		leafparents[MAX_MAP_LEAFS];
int		nodeparents[MAX_MAP_NODES];

void MakeParents (int nodenum, int parent)
{
	int		i, j;
	dnode_t	*node;

	nodeparents[nodenum] = parent;
	node = &dnodes[nodenum];

	for (i=0 ; i<2 ; i++)
	{
		j = node->children[i];
		if (j < 0)
			leafparents[-j - 1] = nodenum;
		else
			MakeParents (j, nodenum);
	}
}


// Argh! - creates a list of curve normals at each vertex,
//         also sets SURF_CURVE flag on apropriate faces (2.00)
void CalcPhongNormals (void)
{
	int			i, j;
	dface_t		*f;
	int			tag;
	vnorms_t	*vn;
//	double		srcNormal[3];

	memset (&vertnorms, 0, sizeof(vertnorms));	// Knightmare- zero vertnums array

	// go through all faces
	for (i=0; i<numfaces; i++)
	{
		f = &dfaces[i];
		tag = texinfo[f->texinfo].value;	// Knightmare- make sure tag is set

		if ( !( texinfo[f->texinfo].flags & SURF_CURVE ) )
		{
			// if not phong shaded, move on
			// Knightmare- don't set tag here
		//	if (!(tag = texinfo[f->texinfo].value) ||
			if ( !(tag)
				|| ((texinfo[f->texinfo].flags & SURF_LIGHT) || (texinfo[f->texinfo].flags & SURF_FOGPLANE)) )
				continue;
		}

		// set new curve flag on face
		texinfo[f->texinfo].flags |= SURF_CURVE;

		// go through all face vertices
		for (j=0; j<f->numedges; j++)
		{
			int e, p;

			e = dsurfedges[f->firstedge + j];

			if (e >= 0)
				p = dedges[e].v[0];
			else
				p = dedges[-e].v[1];

			// look for previous vertnorms entry for this curve
			for (vn=vertnorms[p]; vn; vn=vn->next)
			{
				if (vn->tag == tag)
					break;
			}

			// if no previous entry, create new one
			if (!vn)
			{
				vn = malloc(sizeof(vnorms_t));

				vn->tag = tag;
				VectorClear(vn->normal);

				vn->next = vertnorms[p];
				vertnorms[p] = vn;
			}

			// add face normal
			if (f->side)
			{
			//	VectorCopy(backplanes[f->planenum].normal, srcNormal);
			//	VectorAdd(vn->doubleNormal, srcNormal, vn->doubleNormal);
				VectorAdd(vn->normal, backplanes[f->planenum].normal, vn->normal);
			}
			else
			{
			//	VectorCopy(dplanes[f->planenum].normal, srcNormal);
			//	VectorAdd(vn->doubleNormal, srcNormal, vn->doubleNormal);
				VectorAdd(vn->normal, dplanes[f->planenum].normal, vn->normal);
			}
		}
	}

	// normalize all entries
	for (i=0; i<numvertexes; i++)
	{
		for (vn=vertnorms[i]; vn; vn=vn->next) {
		//	VectorCopy(vn->doubleNormal, vn->normal);
			VectorNormalize(vn->normal, vn->normal);
		}
	}
}


/*
===================================================================

TRANSFER SCALES

===================================================================
*/

int	PointInLeafnum (vec3_t point)
{
	int		nodenum;
	vec_t	dist;
	dnode_t	*node;
	dplane_t	*plane;

	nodenum = 0;
	while (nodenum >= 0)
	{
		node = &dnodes[nodenum];
		plane = &dplanes[node->planenum];
		dist = DotProduct (point, plane->normal) - plane->dist;
		if (dist > 0)
			nodenum = node->children[0];
		else
			nodenum = node->children[1];
	}

	return -nodenum - 1;
}


dleaf_t		*PointInLeaf (vec3_t point)
{
	int		num;

	num = PointInLeafnum (point);
	return &dleafs[num];
}


qboolean PvsForOrigin (vec3_t org, byte *pvs)
{
	dleaf_t	*leaf;

	if (!visdatasize)
	{
		memset (pvs, 255, (numleafs+7)/8 );
		return true;
	}

	leaf = PointInLeaf (org);
	if (leaf->cluster == -1)
		return false;		// in solid leaf

	DecompressVis (dvisdata + dvis->bitofs[leaf->cluster][DVIS_PVS], pvs);
	return true;
}

int	total_transfer;

void MakeTransfers (int i)
{
	int			j;
	vec3_t		delta;
	vec_t		dist, scale;
	float		trans;
	int			itrans;
	patch_t		*patch, *patch2;
	float		total;
	dplane_t	plane;
	vec3_t		origin;
	float		transfers[MAX_PATCHES], *all_transfers;
	int			s;
	int			itotal;
	byte		pvs[(MAX_MAP_LEAFS+7)/8];
	int			cluster;

	patch = patches + i;
	total = 0;

	VectorCopy (patch->origin, origin);
	plane = *patch->plane;

	if (!PvsForOrigin (patch->origin, pvs))
		return;

	// find out which patch2s will collect light
	// from patch

	all_transfers = transfers;
	patch->numtransfers = 0;
	for (j=0, patch2 = patches ; j<num_patches ; j++, patch2++)
	{
		transfers[j] = 0;

		if (j == i)
			continue;

		// check pvs bit
		if (!nopvs)
		{
			cluster = patch2->cluster;
			if (cluster == -1)
				continue;
			if ( ! ( pvs[cluster>>3] & (1<<(cluster&7)) ) )
				continue;		// not in pvs
		}

		// calculate vector
		VectorSubtract (patch2->origin, origin, delta);
		dist = VectorNormalize (delta, delta);
		if (!dist)
			continue;	// should never happen

		// reletive angles
		// Argh! - use stored patch normals (2.00)
		scale = DotProduct (delta, patch->normal);
		scale *= -DotProduct (delta, patch2->normal);
		if (scale <= 0)
			continue;

		trans = scale * patch2->area / (dist*dist);

		// Argh! - replaces <0 check, rejects transfers below radmin,
		//         faster check so moved above testline (1.50)
		if (trans < radmin)
			continue;

		// check exact transfer
		if (TestLine_r (0, patch->origin, patch2->origin))
			continue;

		transfers[j] = trans;
		if (trans > 0)
		{
			total += trans;
			patch->numtransfers++;
		}
	}

	// copy the transfers out and normalize
	// total should be somewhere near PI if everything went right
	// because partial occlusion isn't accounted for, and nearby
	// patches have underestimated form factors, it will usually
	// be higher than PI
	if (patch->numtransfers)
	{
		transfer_t	*t;

		if (patch->numtransfers < 0 || patch->numtransfers > MAX_PATCHES)
			Error ("Weird numtransfers");
		s = patch->numtransfers * sizeof(transfer_t);
		patch->transfers = malloc (s);
		if (!patch->transfers)
			Error ("Memory allocation failure");

		//
		// normalize all transfers so all of the light
		// is transfered to the surroundings
		//
		t = patch->transfers;
		itotal = 0;
		for (j=0 ; j<num_patches ; j++)
		{
			if (transfers[j] <= 0)
				continue;
			itrans = transfers[j]*0x10000 / total;
			itotal += itrans;
			t->transfer = itrans;
			t->patch = j;
			t++;
		}
	}

	// don't bother locking around this.  not that important.
	total_transfer += patch->numtransfers;
}

void FreeTransfers (void)
{
	int		i;

	for (i=0 ; i<num_patches ; i++)
	{
		free (patches[i].transfers);
		patches[i].transfers = NULL;
	}
}

void WriteWorld (char *name)
{
	int		i, j;
	FILE		*out;
	patch_t		*patch;
	winding_t	*w;

	out = fopen (name, "w");
	if (!out)
		Error ("Couldn't open %s", name);

	for (j=0, patch=patches ; j<num_patches ; j++, patch++)
	{
		w = patch->winding;
		fprintf (out, "%i\n", w->numpoints);
		for (i=0 ; i<w->numpoints ; i++)
		{
			fprintf (out, "%5.2f %5.2f %5.2f %5.3f %5.3f %5.3f\n",
				w->p[i][0],
				w->p[i][1],
				w->p[i][2],
				patch->totallight[0],
				patch->totallight[1],
				patch->totallight[2]);
		}
		fprintf (out, "\n");
	}

	fclose (out);
}

void WriteGlView (void)
{
	char	name[1024];
	FILE	*f;
	int		i, j;
	patch_t	*p;
	winding_t	*w;

	strcpy (name, source);
	StripExtension (name);
	strcat (name, ".glr");

	f = fopen (name, "w");
	if (!f)
		Error ("Couldn't open %s", f);

	for (j=0 ; j<num_patches ; j++)
	{
		p = &patches[j];
		w = p->winding;
		fprintf (f, "%i\n", w->numpoints);
		for (i=0 ; i<w->numpoints ; i++)
		{
			fprintf (f, "%5.2f %5.2f %5.2f %5.3f %5.3f %5.3f\n",
				w->p[i][0],
				w->p[i][1],
				w->p[i][2],
				p->totallight[0]/128,
				p->totallight[1]/128,
				p->totallight[2]/128);
		}
		fprintf (f, "\n");
	}

	fclose (f);
}

float CollectLight (void)
{
	int		i, j;
	patch_t	*patch;
	vec_t	total;

	total = 0;

	for (i=0, patch=patches ; i<num_patches ; i++, patch++)
	{
		// skys never collect light, it is just dropped
		if (IsSky(patch->face))
		{
			VectorClear (radiosity[i]);
			VectorClear (illumination[i]);
			continue;
		}

		for (j=0 ; j<3 ; j++)
		{
			patch->totallight[j] += illumination[i][j] / patch->area;
			radiosity[i][j] = illumination[i][j] * patch->reflectivity[j];
		}

		total += radiosity[i][0] + radiosity[i][1] + radiosity[i][2];
		VectorClear (illumination[i]);
	}

	return total;
}


/*
=============
ShootLight

Send light out to other patches
  Run multi-threaded
=============
*/
void ShootLight (int patchnum)
{
	int			k, l;
	transfer_t	*trans;
	int			num;
	patch_t		*patch;
	vec3_t		send;

	// this is the amount of light we are distributing
	// prescale it so that multiplying by the 16 bit
	// transfer values gives a proper output value
	for (k=0 ; k<3 ; k++)
		send[k] = radiosity[patchnum][k] / 0x10000;
	patch = &patches[patchnum];

	trans = patch->transfers;
	num = patch->numtransfers;

	for (k=0 ; k<num ; k++, trans++)
	{
		for (l=0 ; l<3 ; l++)
			illumination[trans->patch][l] += send[l]*trans->transfer;
	}
}

void BounceLight (void)
{
	int		i, j;
	float	added;
	char	name[64];
	patch_t	*p;

	for (i=0 ; i<num_patches ; i++)
	{
		p = &patches[i];
		for (j=0 ; j<3 ; j++)
		{
			radiosity[i][j] = p->samplelight[j] * p->reflectivity[j] * p->area;
		}
	}

	for (i=0 ; i<numbounce ; i++)
	{
		RunThreadsOnIndividual (num_patches, false, ShootLight);
		added = CollectLight ();

		qprintf ("bounce:%i added:%f\n", i, added);
		if ( dumppatches && (i==0 || i == numbounce-1) )
		{
			sprintf (name, "bounce%i.txt", i);
			WriteWorld (name);
		}
	}
}



//==============================================================

void CheckPatches (void)
{
	int		i;
	patch_t	*patch;

	for (i=0 ; i<num_patches ; i++)
	{
		patch = &patches[i];

		// Argh! - just zero the negative patches
		if (patch->totallight[0] < 0)
			patch->totallight[0] = 0;
		if (patch->totallight[1] < 0)
			patch->totallight[1] = 0;
		if (patch->totallight[2] < 0)
			patch->totallight[2] = 0;
	}
}

void RadWorld (void)
{
	if (numnodes == 0 || numfaces == 0)
		Error ("Empty map");
	MakeBackplanes ();
	MakeParents (0, -1);
	MakeTnodes (&dmodels[0]);

	// Argh! - pre-calculate vertex phong normals (2.00)
	if (!nocurve)
		CalcPhongNormals ();

	// turn each face into a single patch
	MakePatches ();

	// subdivide patches to a maximum dimension
	SubdividePatches ();

	// create directlights out of patches and lights
	CreateDirectLights ();

	// build initial facelights
	RunThreadsOnIndividual (numfaces, true, BuildFacelights);

	// Argh! - from SiN qrad3 util (1.50)
	FreeDirectLights ();

	if (numbounce > 0)
	{
		// build transfer lists
		RunThreadsOnIndividual (num_patches, true, MakeTransfers);
		qprintf ("transfer lists: %5.1f megs\n"
		, (float)total_transfer * sizeof(transfer_t) / (1024*1024));

		// spread light around
		BounceLight ();

		FreeTransfers ();

		CheckPatches ();
	}

	if (glview)
		WriteGlView ();

	// blend bounced light into direct light and save

	lightdatasize = 0;
	RunThreadsOnIndividual (numfaces, true, FinalLightFace);
}


/*
========
main

light modelfile
========
*/
int main (int argc, char **argv)
{
	int		i;
	double		start, end;
	char		name[1024];
	// Argh!
	char		usage[1024];
	char		logfile[1024]; // (dk)
	char		*env;          // (dk)

	// Argh! - init settings
	ambient[0] = 0.0;
	ambient[1] = 0.0;
	ambient[2] = 0.0;

	verbose = false;

	sprintf(gamedir, "");
	sprintf(moddir, "_");

	// Argh! - set usage string
	strcpy(usage, "----- Usage -----\n");
	strcat(usage, "QRAD.EXE  [-?|help] [-v] [-dump] [-threads num] [-glview] [-nopvs] [-extra]\n");
	strcat(usage, " [-bounce num] [-nobmodlight] [-nocurve] [-nosplotchfix] [-nostopbleed]\n");
	strcat(usage, " [-chop num] [-choplight num] [-chopsky num] [-chopwarp num] [-chopcurve num]\n");
	strcat(usage, " [-scale num] [-direct num] [-entity num] [-onlybounce] [-nocolor|nocoloru]\n");
	strcat(usage, " [-ambient|ambienta|minlight|minlighta num] [-maxlight|maxlighta num]\n");
	strcat(usage, " [-nominlimit] [-gamedir path] [-moddir path]  BSPFILE.BSP\n");

	// Argh! - header and improved settings output, comments
	//         signify the ArghRad additions (rev2.01)
	printf ("----- ArghRad 2.02 by Tim Wright (Argh!) -----\n");
	printf ("Modified from original source code by id Software\n\n");

	printf ("----- Settings -----\n");
	for (i=1 ; i<argc ; i++)
	{
		if (!strcmp(argv[i],"-?") ||
			!strcmp(argv[i],"-help")) // usage
		{
			printf(usage);
			exit(0);
		}
		else if (!strcmp(argv[i],"-argh")) // easter egg! =)
		{
			printf("\n");
			printf("                 dHH                   dHP      dH\n");
			printf("                dHHH                  dH        HH\n");
			printf("               dP HH   dHdHP  dHHHHH  HH dHHh   HH\n");
			printf("           dHHHHHHHH  dHH    dH  dHH  HHdP HH   \"H\n");
			printf("           P dP   HH  PHH    HH dPHH  HHP  HHdP  P\n");
			printf("            dP    HH   HHP   \"HHP HP  HH   HHP\n");
			printf("          dHP     \"HHh HP   n    dP   HP   HP   dP\n");
			printf("                            \"HHHHP\n");
			printf("\n");
			exit(0);
		}
		else if (!strcmp(argv[i],"-dump"))
		{
			dumppatches = true;
			printf(" -dump enabled  (patch info dump)\n");
		}
		else if (!strcmp(argv[i],"-bounce"))
		{
			numbounce = atoi (argv[i+1]);
			i++;
			printf(" -bounce set to %d  (radiosity bounces)\n", numbounce);
		}
		else if (!strcmp(argv[i],"-v"))
		{
			verbose = true;
			printf(" -verbose enabled  (more detailed output)\n");
		}
		else if (!strcmp(argv[i],"-l")) // (dk)
		{
			verbose = true;
			printf ("-verbose enabled\n");
			logging = true;
			printf ("-logging enabled\n");
		}
		else if (!strcmp(argv[i],"-extra"))
		{
			extrasamples = true;
			printf (" -extra enabled  (extra quality light sampling)\n");
		}
		else if (!strcmp(argv[i],"-threads"))
		{
			numthreads = atoi (argv[i+1]);
			i++;
			printf(" -threads set to %d  (multiple CPU processes)\n", numthreads);
		}
		else if (!strcmp(argv[i],"-chop"))
		{
			subdiv = atof (argv[i+1]);
			i++;
			printf(" -chop set to %.1f  (surface patch size)\n", subdiv);
		}
		else if (!strcmp(argv[i],"-chopsky")) // special sky subdiv
		{
			subdivsky = atof (argv[i+1]);
			i++;
			printf(" -chopsky set to %.1f  (sky patch size)\n", subdivsky); // (fixed1.01)
		}
		else if (!strcmp(argv[i],"-chopwarp")) // special warp subdiv
		{
			subdivwarp = atof (argv[i+1]);
			i++;
			printf(" -chopwarp set to %.1f  (warping patch size)\n", subdivwarp); // (fixed1.01)
		}
		else if (!strcmp(argv[i],"-choplight")) // special light subdiv
		{
			subdivlight = atof (argv[i+1]);
			i++;
			printf(" -choplight set to %.1f  (normal light patch size)\n", subdivlight); // (fixed1.01)
		}
		else if (!strcmp(argv[i],"-chopcurve")) // special curve subdiv (2.00)
		{
			subdivcurve = atof (argv[i+1]);
			i++;
			printf(" -chopcurve set to %.1f  (curve surface patch size)\n", subdivcurve); // (2.00))
		}
		else if (!strcmp(argv[i],"-scale"))
		{
			lightscale = atof (argv[i+1]);
			i++;
			printf(" -scale set to %.4f  (lighting brightness scale)\n", lightscale);
		}
		else if (!strcmp(argv[i],"-direct"))
		{
			direct_scale *= atof(argv[i+1]);
			printf (" -direct set to %.4f  (surface lighting brightness scale)\n", direct_scale);
			i++;
		}
		else if (!strcmp(argv[i],"-entity"))
		{
			entity_scale *= atof(argv[i+1]);
			printf (" -entity set to %.4f  (entity lighting brightness scale)\n", entity_scale);
			i++;
		}
		else if (!strcmp(argv[i],"-glview"))
		{
			glview = true;
			printf (" -glview enabled  (?)\n");
		}
		else if (!strcmp(argv[i],"-nopvs"))
		{
			nopvs = true;
			printf (" -nopvs enabled  (stop pvs checking)\n");
		}
		else if (!strcmp(argv[i],"-nocolor")) // RGB weighted nocolor
		{
			nocolor = 1;
			printf (" -nocolor enabled  (lighting converted to greyscale, RGB weighted)\n");
		}
		else if (!strcmp(argv[i],"-nocoloru")) // unweighted nocolor
		{
			nocolor = 2;
			printf (" -nocoloru enabled  (lighting converted to greyscale, unweighted)\n");
		}
		else if (!strcmp(argv[i],"-ambient") || !strcmp(argv[i],"-minlight"))
		{
			ambient[0] = atof(argv[i+1]) * 128;
			ambient[1] = ambient[0];
			ambient[2] = ambient[0];
			i++;
			printf(" -ambienta set to %.1f  (global ambient brightness)\n", ambient[0]);
		}
		else if (!strcmp(argv[i],"-ambienta") || !strcmp(argv[i],"-minlighta"))
		{
			ambient[0] = atof(argv[i+1]); // uses absolute scale
			ambient[1] = ambient[0];
			ambient[2] = ambient[0];
			i++;
			printf(" -ambienta set to %.1f  (global ambient brightness)\n", ambient[0]);
		}
		else if (!strcmp(argv[i],"-maxlight"))
		{
			maxlight = atof(argv[i+1]) * 128;
			i++;
			printf(" -maxlighta set to %.1f  (maximum lighting brightness)\n", maxlight);
		}
		else if (!strcmp(argv[i],"-maxlighta"))
		{
			maxlight = atof(argv[i+1]); // uses absolute scale
			i++;
			printf(" -maxlighta set to %.1f  (maximum lighting brightness)\n", maxlight);
		}
		else if (!strcmp(argv[i],"-radmin")) // min transfer cutoff (1.50)
		{
			radmin = atof(argv[i+1]);
			i++;
			printf(" -radmin set to %.4f  (radiosity minimum cutoff)\n", radmin);
		}
		else if (!strcmp(argv[i],"-nominlimit")) // antilights can go < minlight
		{
			minlimit = false;
			printf (" -nominlimit enabled  (antilights go darker than minlight)\n");
		}
		else if (!strcmp(argv[i],"-nobmodlight")) // bmodels can't emit light
		{
			bmodlight = false;
			printf(" -nobmodlight enabled  (brush model lighting disabled)\n"); // (fixed1.01)
		}
		else if (!strcmp(argv[i],"-nostopbleed")) // light may bleed through walls (1.50)
		{
			stopbleed = false;
			printf(" -nostopbleed enabled  (bleeding light correction disabled)\n");
		}
		else if (!strcmp(argv[i],"-nocurve")) // disables phong shading (2.00)
		{
			nocurve = true;
			printf(" -nocurve enabled  (phong shading disabled)\n");
		}
		else if (!strcmp(argv[i],"-nosplotchfix")) // disables anti-splotch fix (2.01)
		{
			splotchfix = false;
			printf(" -nosplotchfix enabled  (anti-splotch fix disabled for plain face lights)\n");
		}
		else if (!strcmp(argv[i],"-gamedir")) // set gamedir
		{
			sprintf(gamedir, "%s", argv[i+1]);
			i++;
			printf(" -gamedir set to %s  (game pak dir)\n", gamedir);
		}
		else if (!strcmp(argv[i],"-moddir")) // set moddir
		{
			sprintf(moddir, "%s", argv[i+1]);
			i++;
			printf(" -moddir set to %s  (mod pak dir)\n", moddir);
		}
		else if (!strcmp(argv[i],"-onlybounce")) // only bounced light saved (2.00)
		{
			onlybounce = true;
			printf (" -onlybounce enabled  (only bounced light saved)\n");
		}
		else if (!strcmp (argv[i],"-tmpin"))
			strcpy (inbase, "/tmp");
		else if (!strcmp (argv[i],"-tmpout"))
			strcpy (outbase, "/tmp");
		else
			break;
	}

	// (dk)
	if (logging)
	{
		strcpy (logfile, ExpandArg(argv[i]));
		StripExtension (logfile);
		DefaultExtension (logfile, ".rlg");
		OpenLog (logfile);
	}

	ThreadSetDefault ();

	if (maxlight > 255)
		maxlight = 255;

	// Argh! - modded checks to allow subdiv* > subdiv (rev2.01)
	if (subdiv > 256)
		subdiv = 256;
	if (subdivlight == 0)
		subdivlight = subdiv;
	if (subdivsky == 0)
		subdivsky = subdiv;
	if (subdivwarp == 0)
		subdivwarp = subdiv;
	if (subdivcurve == 0)
		subdivcurve = __min(subdiv, 32);

	if (i != argc - 1)
		Error (usage);

	start = I_FloatTime ();

	// Argh! - try environment var first, then path (dk)
	if (!gamedir[0])
	{
		env = getenv ("GAMEDIR");	// Knightmare- set env first
//		if (env = getenv ("GAMEDIR"))
		if (env)
			strcpy(gamedir, env);
	}
	if (!gamedir[0])
		SetQdirFromPath (argv[i]);
	if (!gamedir[0])
		Error ("Could not get gamedir: set environment var GAMEDIR, or use -gamedir");

	// Argh! - needs trailing slash (dk)
	if (gamedir[strlen(gamedir)-1] != '\\' && gamedir[strlen(gamedir)-1] != '/')
		strcat(gamedir, "\\");
	if (moddir[strlen(moddir)-1] != '\\' && moddir[strlen(moddir)-1] != '/')	// Knightmare 9/4/12- fix use of strlen(gamedir)
		strcat(moddir, "\\");

	strcpy (source, ExpandArg(argv[i]));
	StripExtension (source);
	DefaultExtension (source, ".bsp");

//	ReadLightFile ();

	printf("----- Load BSP File -----\n");
	sprintf (name, "%s%s", inbase, source);
	printf ("reading %s\n", name);
	LoadBSPFile (name);
	ParseEntities ();
	CalcTextureReflectivity ();

	if (!visdatasize)
	{
		printf ("no vis information, direct lighting only\n");
		numbounce = 0;
		// Argh! - removed ambient value (1.50)
	}

	printf("----- Light World -----\n");
	RadWorld ();

	printf("----- Save BSP File -----\n");
	sprintf (name, "%s%s", outbase, source);
	printf ("writing %s\n", name);
	WriteBSPFile (name);

	printf("----- Time -----\n");
	end = I_FloatTime ();
	printf ("%5.0f seconds elapsed\n", end-start);

	if (logging) // (dk)
		CloseLog ();

	return 0;
}

